import logging
import os
import tempfile
from typing import TYPE_CHECKING

from strong_opx.exceptions import CommandError
from strong_opx.template import FileTemplate

if TYPE_CHECKING:
    from strong_opx.platforms import Platform
    from strong_opx.project import Environment, Project

logger = logging.getLogger(__name__)


class NodeConfig(str):
    def __new__(cls, name: str, path: str):
        instance = super().__new__(cls, name)
        instance.path = path
        return instance


class DeploymentProvider:
    name: str
    platform: "Platform"

    def __init__(self, project: "Project", environment: "Environment", platform: "Platform"):
        self.project = project
        self.environment = environment
        self.platform = platform

    def select_nodes(self, nodes: list[str]) -> list[str]:
        prefix = f"{self.name}/"
        return [node for node in nodes if node.startswith(prefix)]

    def deploy_node(self, node: str) -> bool:
        raise NotImplementedError()


class ConfigDeploymentProvider(DeploymentProvider):
    allowed_extensions: tuple[str, ...] = None

    def select_nodes(self, nodes: list[str]) -> list[NodeConfig]:
        provide_nodes = []
        for node in nodes:
            if node.startswith(f"{self.name}/"):
                node_path = os.path.join(self.project.path, node)
            elif node.startswith(f"{self.name}:"):
                _, node_path = node.split(":", 1)
            else:
                continue

            if not os.path.exists(node_path):
                raise CommandError(f"Unable to locate {node} at {node_path}")

            provide_nodes.append(NodeConfig(node, node_path))

        return provide_nodes

    def is_config_applicable(self, file_name: str) -> bool:
        file_name, extension = os.path.splitext(file_name)
        if self.allowed_extensions and extension not in self.allowed_extensions:
            return False

        environment = os.path.splitext(file_name)[1]
        return not environment or environment.replace(".", "") == self.environment.name

    def deploy_node(self, node: NodeConfig) -> bool:
        if os.path.isdir(node.path):
            config_files = [
                os.path.join(node.path, file_name)
                for file_name in os.listdir(node.path)
                if self.is_config_applicable(file_name=file_name)
            ]
        else:
            config_files = [node.path]

        if not config_files:  # Empty directory or no config for this environment
            logger.warning("No config file found for current environment. Deployment skipped")
            return False

        with tempfile.TemporaryDirectory(prefix=os.path.basename(node.path)) as td:
            for source_path in config_files:
                target_path = os.path.join(td, os.path.basename(source_path))
                FileTemplate(source_path).render_to_file(target_path, self.environment.context)

            return self.deploy(node, td)

    def deploy(self, node: NodeConfig, directory: str) -> bool:
        raise NotImplementedError()



import argparse
import base64
import json
import tempfile
import threading
import time
import webbrowser
from typing import TYPE_CHECKING

from strong_opx.exceptions import PluginError, ProcessError
from strong_opx.platforms.plugins.plugin import PlatformPlugin

if TYPE_CHECKING:
    from strong_opx.platforms.kubernetes import KubernetesPlatform

EKS_ADMIN_NAME = "eks-admin"
EKS_ADMIN_SERVICE_ACCOUNT = f"""
apiVersion: v1
kind: ServiceAccount
metadata:
  name: {EKS_ADMIN_NAME}
  namespace: kube-system
---
apiVersion: rbac.authorization.k8s.io/v1
kind: ClusterRoleBinding
metadata:
  name: {EKS_ADMIN_NAME}
roleRef:
  apiGroup: rbac.authorization.k8s.io
  kind: ClusterRole
  name: cluster-admin
subjects:
- kind: ServiceAccount
  name: {EKS_ADMIN_NAME}
  namespace: kube-system
  """


class KubernetesDashboardPlugin(PlatformPlugin):
    platform: "KubernetesPlatform"
    config_url = "https://raw.githubusercontent.com/kubernetes/dashboard/v2.0.5/aio/deploy/recommended.yaml"

    def is_installed(self) -> bool:
        output = self.platform.kubectl(
            "get",
            "service",
            "kubernetes-dashboard",
            "-n",
            "kubernetes-dashboard",
            capture_output=True,
            ignore_exit_code=True,
        )
        if output.returncode:
            if b"(NotFound)" in output.stderr:
                return False

            raise ProcessError(output.stderr)

        return True

    def install(self) -> None:
        self.platform.kubectl("apply", "-f", self.config_url)

        with tempfile.NamedTemporaryFile(suffix=".yml", mode="w+") as f:
            f.write(EKS_ADMIN_SERVICE_ACCOUNT)
            f.flush()

            self.platform.kubectl("apply", "-f", f.name)

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        subparsers = parser.add_subparsers(title="operation", description="Operation to execute", required=True)

        up_parser = subparsers.add_parser("up", help="Start Kubernetes Dashboard")
        up_parser.add_argument("-d", "--detach", action="store_true", dest="detached", default=False)
        up_parser.add_argument("--no-browser", action="store_false", dest="open_browser", default=True)
        up_parser.set_defaults(operation="up")

        down_parser = subparsers.add_parser("down", help="Stop Kubernetes Dashboard")
        down_parser.set_defaults(operation="down")

        token_parser = subparsers.add_parser("token", help="Get Kubernetes Dashboard auth token")
        token_parser.set_defaults(operation="token")

    def up(self, open_browser: bool, detached: bool):
        def delayed_open_browser():
            port = None
            time.sleep(1)
            while True:
                status = self.platform.get_proxy_status()
                if status is None:
                    time.sleep(1)
                    continue

                _, port = status
                break

            time.sleep(1)
            webbrowser.open(
                f"http://localhost:{port}/api/v1/namespaces/kubernetes-dashboard/services"
                f"/https:kubernetes-dashboard:/proxy/#!/login",
                new=2,
            )

        open_browser_thread = None
        if open_browser:
            open_browser_thread = threading.Thread(target=delayed_open_browser)
            open_browser_thread.start()

        self.platform.start_proxy(detached)
        if open_browser_thread is not None:
            open_browser_thread.join()

    def get_secret_name(self) -> str:
        output = self.platform.kubectl("-n=kube-system", "get", "secret", "-o=name", capture_output=True).stdout
        matching_lines = [line for line in output.decode("utf8").split() if EKS_ADMIN_NAME in line]
        if not matching_lines:
            raise PluginError(f"Unable to locate secret for {EKS_ADMIN_NAME}")

        return matching_lines[0]

    def token(self):
        self.platform.ensure_proxy_is_running()
        secret_name = self.get_secret_name()

        output = self.platform.kubectl("-n=kube-system", "get", "-o=json", secret_name, capture_output=True)

        secret_detail = json.loads(output.stdout.decode("utf8"))
        secret_token = secret_detail["data"]["token"]
        secret_token = base64.b64decode(secret_token).decode("utf8")

        print("\nUse the following token to log into the Kubernetes Dashboard:\n")
        print(secret_token)
        print()

    def handle(self, operation: str, **kwargs):
        if operation == "up":
            self.up(**kwargs)
        elif operation == "down":
            self.platform.stop_proxy()
        elif operation == "token":
            self.token()



import base64

from strong_opx.template.template import Template


@Template.register_filter("base64")
def base64_filter(v: str) -> str:
    return base64.b64encode(v.encode("utf-8")).decode("utf-8")



from typing import Optional

import boto3


class FilterModule:
    def __init__(self):
        self.client = boto3.client("ec2")

    def filters(self):
        return {
            "security_group_id": self.security_group_id,
            "ec2_public_ip": self.public_ip,
            "ec2_private_ip": self.private_ip,
            "vpc_id": self.vpc_id,
        }

    def security_group_id(self, sg_name: str, vpc_id: Optional[str]) -> Optional[str]:
        filters = [{"Name": "group-name", "Values": [sg_name]}]
        if vpc_id:
            filters.append({"Name": "vpc-id", "Values": [vpc_id]})

        response = self.client.describe_security_groups(Filters=filters)["SecurityGroups"]
        if len(response) == 1:
            return response[0]["GroupId"]

    def describe_instance(self, instance_name: str) -> dict:
        reservations = self.client.describe_compute_instance(Filters=[{"Name": "tag:Name", "Values": [instance_name]}])[
            "Reservations"
        ]

        instances = []
        for reservation in reservations:
            instances.extend(reservation["Instances"])

        if len(instances) == 0:
            raise ValueError(f"Unable to find instance with name: {instance_name}")

        if len(instances) > 1:
            raise ValueError(f"Found multiple instances with name: {instance_name}")

        return instances[0]

    def public_ip(self, instance_name: str) -> str:
        instance = self.describe_instance(instance_name)
        ip_address = instance.get("PublicIpAddress")
        if not ip_address:
            raise ValueError(f"Instance {instance_name} does not have a public IP")

        return ip_address

    def private_ip(self, instance_name: str) -> str:
        instance = self.describe_instance(instance_name)
        return instance["PrivateIpAddress"]

    def vpc_id(self, vpc_name: str, cidr_block: str = None) -> str:
        filters = [{"Name": "tag:Name", "Values": [vpc_name]}]
        if cidr_block:
            filters.append({"Name": "cidr-block-association.cidr-block", "Values": [cidr_block]})

        vpcs = self.client.describe_vpcs(Filters=filters, MaxResults=2)["Vpcs"]
        if len(vpcs) == 1:
            return vpcs[0]["VpcId"]

        raise ValueError(f"Found multiple VPCs with name: {vpc_name}. Specify a CIDR block to disambiguate")



import importlib
import pkgutil
import warnings
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar

from colorama import Style

from strong_opx.exceptions import ConfigurationError, ErrorDetail, ImproperlyConfiguredError
from strong_opx.utils.module_loading import import_string
from strong_opx.utils.tracking import get_position

if TYPE_CHECKING:
    from strong_opx.providers import Provider

T = TypeVar("T")

_selected_provider_name: Optional[str] = None


@lru_cache()
def known_providers() -> tuple[str, ...]:
    root_module = importlib.import_module("strong_opx.providers")
    root_module_path = root_module.__path__

    providers = [name for _, name, is_package in pkgutil.iter_modules(root_module_path) if is_package]

    return tuple(providers)


def get_provider_class(provider_name: str) -> type["Provider"]:
    try:
        module = importlib.import_module(f"strong_opx.providers.{provider_name}")
    except ImportError as e:
        raise ConfigurationError(
            ErrorDetail(
                f"Failed to load {provider_name} provider module because '{e.name}' is not installed",
                hint=(
                    "Check if the strong-opx is installed correctly alongside all required external dependencies.\n\n"
                    f"    {Style.BRIGHT}pip install strong-opx[{provider_name}]{Style.RESET_ALL}"
                ),
                *get_position(provider_name),
            )
        )

    provider_class_name = getattr(module, "provider_class")
    provider_class = import_string(provider_class_name)
    setattr(provider_class, "name", provider_name)
    return provider_class


def select_provider(config_dict: dict[str, Any]) -> tuple[str, type["Provider"]]:
    found_providers = []

    # We need to keep tracking info attached with config_keys to provide better error messages.
    # While rewriting make sure tracking info is not lost.
    for config_key in config_dict.keys():
        if config_key in known_providers():
            found_providers.append(config_key)

    if len(found_providers) == 0:
        warnings.warn(
            "No provider is specified. Defaulting to `aws`. Include `aws:` in strong-opx.yml to suppress this warning"
        )
        provider_name = "aws"
    elif len(found_providers) > 1:
        raise ImproperlyConfiguredError("Multiple providers are specified. This is not allowed.")
    else:
        provider_name = found_providers.pop()

    # Replace None with empty dictionary
    if config_dict.get(provider_name) is None:
        config_dict[provider_name] = {}

    global _selected_provider_name
    _selected_provider_name = provider_name

    return provider_name, get_provider_class(provider_name)


def current_provider_name() -> str:
    if _selected_provider_name is None:
        raise RuntimeError("`current_provider_name` is called before `select_provider`")

    return _selected_provider_name


@lru_cache()
def current_provider_class() -> Type["Provider"]:
    return get_provider_class(current_provider_name())


@lru_cache()
def current_provider() -> "Provider":
    from strong_opx.project import Project

    return Project.current().provider


def current_provider_error_handler(ex: Exception) -> None:
    from strong_opx.project import Project

    project = Project.current()
    if project is None:
        raise ex

    project.provider.handle_error(ex)



import contextlib
import os
import shlex
import subprocess
import sys

from colorama import Fore as ConsoleForeground
from colorama import Style

from strong_opx.exceptions import ProcessError


def shell(command, **kwargs) -> subprocess.CompletedProcess:
    ignore_exit_code = kwargs.pop("ignore_exit_code", False)

    if isinstance(command, str):
        print(Style.DIM, f"$ {command}", Style.RESET_ALL, sep="")
    else:
        str_command = " ".join(shlex.quote(c) if " " in c else c for c in command)
        print(Style.DIM, f"$ {str_command}", Style.RESET_ALL, sep="")

    try:
        results = subprocess.run(command, **kwargs)
        if not ignore_exit_code and results.returncode:
            message = f"Exit Code: {results.returncode}\n"

            if kwargs.get("capture_output", False):
                if results.stdout:
                    message += "\n<< Captured Output >>\n"
                    message += results.stdout.decode("utf8").strip()
                    message += "\n"

                if results.stderr:
                    message += f"\n<< Captured Error >>\n{ConsoleForeground.RED}"
                    message += results.stderr.decode("utf8").strip()
                    message += f"{Style.RESET_ALL}"

            raise ProcessError(message.strip())

        return results
    except KeyboardInterrupt:
        print("--KeyboardInterrupt--", file=sys.stderr)
        exit(1)


def static_eval_bash_vars(vars_str: str) -> dict[str, str]:
    bash_vars = {}
    for statement in vars_str.split(";"):
        statement = statement.strip()
        parts = statement.split("=")
        if len(parts) != 2:
            continue

        name, value = parts
        if name.startswith("export "):
            name = name[7:]

        if " " not in name and name.isupper() and value[0] != " ":
            bash_vars[name] = value

    return bash_vars


@contextlib.contextmanager
def ssh_agent(ssh_key: str = None):
    stdout = shell("ssh-agent", capture_output=True).stdout.decode("utf8")
    ssh_agent_info = static_eval_bash_vars(stdout)
    pid = ssh_agent_info.get("SSH_AGENT_PID")
    if not pid or not pid.isdigit():
        raise ProcessError(f"Failed to start ssh-agent; Output: {stdout}")

    try:
        for k, v in ssh_agent_info.items():
            os.environ[k] = v

        if ssh_key:
            shell(["ssh-add", ssh_key])
        else:
            shell(["ssh-add", "-K"])

        yield
    finally:
        for k in ssh_agent_info:
            os.environ.pop(k, None)

        os.kill(int(pid), 9)



import tempfile
from unittest import TestCase
from unittest.mock import MagicMock, patch

from strong_opx.exceptions import ErrorDetail
from strong_opx.utils.tracking import Position
from tests.helper_functions import patch_colorama


def new_error_detail(start_line: int, start_column: int, end_line: int = None, end_column: int = None) -> ErrorDetail:
    start_pos = Position(start_line, start_column)

    end_pos = None
    if end_line is not None and end_column is not None:
        end_pos = Position(end_line, end_column)

    return ErrorDetail(error="", file_path="", start_pos=start_pos, end_pos=end_pos)


@patch_colorama
class ErrorDetailTests(TestCase):
    def test_read_block__single_line(self):
        with tempfile.NamedTemporaryFile() as f:
            for i in range(1, 10):
                f.write(f"line {i}\n".encode("utf8"))

            f.flush()
            e = new_error_detail(1, 3, 1, 7)
            block = list(e.read_block(f.name))

        self.assertListEqual(
            [
                "{Style.DIM}1:  {Style.RESET_ALL}li{Style.BRIGHT}ne 1{Style.RESET_ALL}",
                "{Style.DIM}2:  {Style.RESET_ALL}line 2",
                "{Style.DIM}3:  {Style.RESET_ALL}line 3",
            ],
            block,
        )

    def test_read_block__multiple_line(self):
        with tempfile.NamedTemporaryFile() as f:
            for i in range(1, 10):
                f.write(f"line {i}\n".encode("utf8"))

            f.flush()
            e = new_error_detail(1, 3, 2, 6)
            block = list(e.read_block(f.name))

        self.assertListEqual(
            [
                "{Style.DIM}1:  {Style.RESET_ALL}li{Style.BRIGHT}ne 1",
                "{Style.DIM}2:  {Style.RESET_ALL}line {Style.RESET_ALL}2",
                "{Style.DIM}3:  {Style.RESET_ALL}line 3",
                "{Style.DIM}4:  {Style.RESET_ALL}line 4",
            ],
            block,
        )

    def test_read_block__remove_context_after_blank_line(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"line 1\n")
            f.write(b"line 2\n")
            f.write(b"line 3\n")
            f.write(b"\n")
            f.write(b"line 5\n")
            f.write(b"line 6\n")
            f.flush()

            block = list(new_error_detail(2, 3, 2, 7).read_block(f.name))

        self.assertListEqual(
            [
                "{Style.DIM}1:  {Style.RESET_ALL}line 1",
                "{Style.DIM}2:  {Style.RESET_ALL}li{Style.BRIGHT}ne 2{Style.RESET_ALL}",
                "{Style.DIM}3:  {Style.RESET_ALL}line 3",
            ],
            block,
        )

    def test_read_block__remove_context_before_blank_line(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"line 1\n")
            f.write(b"\n")
            f.write(b"line 3\n")
            f.write(b"line 4\n")
            f.write(b"line 5\n")
            f.flush()

            block = list(new_error_detail(3, 3, 3, 7).read_block(f.name))

        self.assertListEqual(
            [
                "{Style.DIM}3:  {Style.RESET_ALL}li{Style.BRIGHT}ne 3{Style.RESET_ALL}",
                "{Style.DIM}4:  {Style.RESET_ALL}line 4",
                "{Style.DIM}5:  {Style.RESET_ALL}line 5",
            ],
            block,
        )

    def test_read_block__no_end_pos(self):
        """
        Prints the "start" line and `YAMLError.lines_for_context` before and after the start line.
        """
        with tempfile.NamedTemporaryFile() as f:
            for i in range(1, 10):
                f.write(f"line {i}\n".encode("utf8"))

            f.flush()

            block = list(new_error_detail(3, 3).read_block(f.name))

            self.assertListEqual(
                [
                    "{Style.DIM}1:  {Style.RESET_ALL}line 1",
                    "{Style.DIM}2:  {Style.RESET_ALL}line 2",
                    "{Style.DIM}3:  {Style.RESET_ALL}li{Style.BRIGHT}ne 3{Style.RESET_ALL}",
                    "{Style.DIM}4:  {Style.RESET_ALL}line 4",
                    "{Style.DIM}5:  {Style.RESET_ALL}line 5",
                ],
                block,
            )

    def test_read_block__no_end_pos_one_line(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("line 1\n".encode("utf8"))

            f.flush()
            block = list(new_error_detail(0, 3).read_block(f.name))

        self.assertEqual(["{Style.DIM}1:  {Style.RESET_ALL}li{Style.BRIGHT}ne 1{Style.RESET_ALL}"], block)

    def test_str__no_file_path(self):
        e = ErrorDetail("dummy error")
        self.assertEqual("{Fore.RED}Error:{Fore.RESET} {Style.BRIGHT}dummy error{Style.RESET_ALL}", str(e))

    @patch("strong_opx.project.Project.current", new=MagicMock())
    def test_str__with_file_path(self):
        e = ErrorDetail("dummy error", file_path="some-path.yml")
        self.assertEqual(
            "{Fore.RED}╷{Fore.RESET}\n"
            "{Fore.RED}│{Fore.RESET} {Fore.RED}Error:{Fore.RESET} {Style.BRIGHT}dummy error{Style.RESET_ALL}\n"
            "{Fore.RED}│{Fore.RESET} \n"
            "{Fore.RED}│{Fore.RESET}   in some-path.yml\n"
            "{Fore.RED}╵{Fore.RESET}",
            str(e),
        )

    @patch("strong_opx.project.Project.current", new=MagicMock())
    @patch("strong_opx.exceptions.ErrorDetail.read_block")
    def test_str__with_file_path_and_start_pos(self, read_block_mock: MagicMock):
        read_block_mock.side_effect = lambda *_: ["line 1", "line 2"]

        e = ErrorDetail("dummy error", file_path="some-path.yml", start_pos=Position(2, 2))
        self.assertEqual(
            "{Fore.RED}╷{Fore.RESET}\n"
            "{Fore.RED}│{Fore.RESET} {Fore.RED}Error:{Fore.RESET} {Style.BRIGHT}dummy error{Style.RESET_ALL}\n"
            "{Fore.RED}│{Fore.RESET} \n"
            "{Fore.RED}│{Fore.RESET}   in some-path.yml on line 2 column 2\n"
            "{Fore.RED}│{Fore.RESET}   line 1\n"
            "{Fore.RED}│{Fore.RESET}   line 2\n"
            "{Fore.RED}╵{Fore.RESET}",
            str(e),
        )

    def test_str__with_hint(self):
        e = ErrorDetail("dummy error", hint="some cool hint")
        self.assertEqual(
            "{Fore.RED}╷{Fore.RESET}\n"
            "{Fore.RED}│{Fore.RESET} {Fore.RED}Error:{Fore.RESET} {Style.BRIGHT}dummy error{Style.RESET_ALL}\n"
            "{Fore.RED}│{Fore.RESET} \n"
            "{Fore.RED}│{Fore.RESET} some cool hint\n"
            "{Fore.RED}╵{Fore.RESET}",
            str(e),
        )

    def test_str__with_long_hint(self):
        e = ErrorDetail("dummy error", hint="some cool but long hint " * 5)
        self.assertEqual(
            "{Fore.RED}╷{Fore.RESET}\n"
            "{Fore.RED}│{Fore.RESET} {Fore.RED}Error:{Fore.RESET} {Style.BRIGHT}dummy error{Style.RESET_ALL}\n"
            "{Fore.RED}│{Fore.RESET} \n"
            "{Fore.RED}│{Fore.RESET} some cool but long hint some cool but long hint some cool but long hint some\n"
            "{Fore.RED}│{Fore.RESET} cool but long hint some cool but long hint\n"
            "{Fore.RED}╵{Fore.RESET}",
            str(e),
        )



import re
from dataclasses import field
from ipaddress import IPv4Address
from typing import Optional, Union

from pydantic.dataclasses import dataclass
from pydantic_core import CoreSchema, core_schema

from strong_opx.compat import StrEnum
from strong_opx.providers.discovery import current_provider, current_provider_class
from strong_opx.utils.mapping import CaseInsensitiveMultiTagDict


def filter_instances_by_environment_tag_if_exists(
    instances: list["ComputeInstanceDescription"],
) -> list["ComputeInstanceDescription"]:
    """
    Given instances, return all instances which:
    * Do NOT have an "environment" tag (case-insensitive)
    OR
    * DO have an "environment" tag (case-insensitive) AND the value of that tag matches the current
      environment (also case-insensitive)

    This function is intentionally "permissive" with the instances that it does NOT filter out. In particular:
    * Allow instances without an "Environment" tag. Do not assume an instance is not part of the current
      environment, if the instance does not have an "Environment" tag.
    * Allow instances with multiple "Environment" tags, if at least one value matches the current environment. It's
      possible multiple environments use the same instance.

    This "permissiveness" is expected to be more backwards compatible than being more restrictive.
    """

    from strong_opx.project import Project

    selected_environment = Project.current().selected_environment
    this_environment = selected_environment.name.lower()

    def keep_instance(instance: "ComputeInstanceDescription") -> bool:
        environment_tags = instance.tags.get("environment")
        return not environment_tags or any(tag.lower() == this_environment for tag in environment_tags)

    return [instance for instance in instances if keep_instance(instance)]


class ComputeInstance:
    hostname: str

    def __init__(self, value: Union[str, IPv4Address]):
        self.hostname = value
        self._ip_address = None
        self._instance_id = None

        if isinstance(value, IPv4Address):
            self._ip_address = value
        else:
            self._instance_id = value

    @property
    def ip_address(self) -> IPv4Address:
        if self._ip_address is None:
            self._ip_address = self._instance_ip_address()

        return self._ip_address

    @property
    def instance_id(self) -> str:
        if self._instance_id is None:
            self._instance_id = self._instance_id_from_ip_address()

        return self._instance_id

    @property
    def current_state(self) -> "ComputeInstanceState":
        return self.describe().state

    def describe(self):
        if self._instance_id is not None:
            return self._describe_by_instance_id()

        return self._describe_by_ip()

    def _describe_by_ip(self) -> "ComputeInstanceDescription":
        instances = current_provider().query_compute_instances(self.ip_address)

        if self._ip_address.is_private:
            instances = filter_instances_by_environment_tag_if_exists(instances)

        if len(instances) == 1:
            return instances[0]

        public_or_private = "private" if self._ip_address.is_private else "public"

        if len(instances) == 0:
            raise ValueError(f'Unable to find an instance with {public_or_private} IP "{self._ip_address}"')

        raise ValueError(
            f'Expected to find exactly one instance with {public_or_private} IP "{self._ip_address}", but found '
            f"{len(instances)}. "
            f'Instance IDs: {", ".join(instance.instance_id for instance in instances)}.'
        )

    def _describe_by_instance_id(self) -> "ComputeInstanceDescription":
        return current_provider().describe_compute_instance(self.instance_id)

    def _instance_ip_address(self) -> IPv4Address:
        instance = self._describe_by_instance_id()
        if instance.public_ip:
            return IPv4Address(instance)

        return instance.private_ip

    def _instance_id_from_ip_address(self) -> str:
        return self._describe_by_ip().instance_id

    def __str__(self):
        if self._ip_address is not None and self._instance_id is None:
            return str(self._ip_address)

        if self._instance_id is not None and self._ip_address is None:
            return self._instance_id

        return f"{self._ip_address} ({self._instance_id})"

    def __repr__(self):
        return f"<{type(self).__name__}: {self.hostname}>"

    def __eq__(self, other):
        if not isinstance(other, ComputeInstance):
            return False

        return self.hostname == other.hostname

    def __hash__(self):
        return hash(self.hostname)

    @classmethod
    def parse(cls, value: Union[str, IPv4Address]) -> "ComputeInstance":
        if isinstance(value, IPv4Address):
            return cls(value)
        elif not isinstance(value, str):
            raise TypeError(f"Expected str or IPv4Address, got {type(value).__name__}")

        try:
            return cls(IPv4Address(value))
        except ValueError as ex:
            # If it's not a valid IPv4 address, it might be an instance ID
            provider_class = current_provider_class()
            if provider_class.compute_instance_id_re is None:
                raise ex  # If the provider does not support instance IDs, raise the original ValueError

            if not re.match(provider_class.compute_instance_id_re, value):
                raise ValueError(
                    f"Invalid instance ID format: {value}. Expected format: {provider_class.compute_instance_id_re}"
                )

        return cls(value)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler) -> CoreSchema:
        return core_schema.no_info_plain_validator_function(
            cls.parse, ref="ComputeInstance", serialization=core_schema.to_string_ser_schema()
        )


class ComputeInstanceState(StrEnum):
    RUNNING = "running"
    STOPPED = "stopped"
    UNKNOWN = "unknown"


@dataclass
class ComputeInstanceDescription:
    instance_id: str
    state: ComputeInstanceState

    private_ip: IPv4Address
    public_ip: Optional[IPv4Address] = None
    tags: CaseInsensitiveMultiTagDict[str, str] = field(default_factory=CaseInsensitiveMultiTagDict)



import sys
from typing import Collection

from prompt_toolkit.application import Application, get_app
from prompt_toolkit.cursor_shapes import CursorShape
from prompt_toolkit.filters import IsDone
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import Dimension, FormattedTextControl, Layout
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit, Window
from prompt_toolkit.styles import Style
from prompt_toolkit.widgets import RadioList

default_style = Style.from_dict(
    {
        "question-mark": "#5F819D",
        "pointer-selected": "#FF9D00 bold",  # AWS orange
        "answer": "#FF9D00 bold",  # AWS orange
        "question": "bold",
    }
)


class RadioListEx(RadioList):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._total_values = len(self.values)

    def _get_text_fragments(self) -> StyleAndTextTuples:
        tokens = []

        for index, (key, value) in enumerate(self.values):
            selected = index == self._selected_index
            tokens.append(("", " "))
            if selected:
                tokens.append(("[SetCursorPosition]", ""))
                tokens.append(("class:pointer-selected", "\u276f "))
            else:
                tokens.append(("", "  "))

            tokens.append(("class:radio-selected" if selected else "", value))
            tokens.append(("", "\n"))

        tokens.pop()  # Remove last newline.
        return tokens

    @property
    def selected_value(self):
        return self.values[self._selected_index][0]


def create_application(message: str, choices: Collection[str]) -> Application:
    answered = False
    radio_list = RadioListEx(values=list(zip(choices, choices)))

    def get_prompt_tokens():
        tokens = [
            ("class:question-mark", "?"),
            ("class:question", f" {message} "),
        ]

        if answered:
            tokens.append(("class:answer", f" {radio_list.selected_value}"))
        else:
            tokens.append(("", " (Use arrow keys)"))

        return tokens

    # assemble layout
    layout = Layout(
        HSplit(
            [
                Window(
                    height=Dimension.exact(1),
                    content=FormattedTextControl(get_prompt_tokens),
                ),
                ConditionalContainer(
                    HSplit([radio_list]),
                    filter=~IsDone(),
                ),
            ]
        )
    )

    # key bindings
    key_bindings = KeyBindings()

    @key_bindings.add(Keys.ControlQ, eager=True)
    @key_bindings.add(Keys.ControlC, eager=True)
    def _(event):
        get_app().exit(result=None)

    @key_bindings.add(Keys.Enter, eager=True)
    def set_answer(event):
        nonlocal answered
        answered = True

        get_app().exit(result=radio_list.selected_value)

    return Application(layout=layout, key_bindings=key_bindings, style=default_style, cursor=CursorShape.BLOCK)


def select_prompt(message: str, choices: Collection[str]) -> str:
    application = create_application(message, choices)
    selection = application.run()
    if selection is None:
        print("--KeyboardInterrupt--", file=sys.stderr)
        exit(1)

    return selection


def input_boolean(prompt, default=True) -> bool:
    if default is None:
        hint = "[y/n]"
    elif default:
        hint = "[Y/n] "
    else:
        hint = "[y/N] "

    while True:
        choice = input(f"{prompt} {hint}: ").strip().lower()
        if default is not None and choice == "":
            return default

        if choice == "y":
            return True

        if choice == "n":
            return False

        print('Please respond with "y" or "n")')



import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from strong_opx.project import Project
    from strong_opx.template import Context


class EnvironHook:
    def __init__(self, prefix: str):
        if not prefix:
            raise ValueError("prefix cannot be blank")

        self.prefix = prefix

    def __call__(self, context: "Context"):
        for k in os.environ:
            if k.startswith(self.prefix):
                context[k] = os.environ[k]


class ProjectContextHook:
    def __init__(self, project: "Project"):
        self.project = project

    def __call__(self, context: "Context"):
        context["SSH_KEY"] = lambda: self.project.config.ssh_key
        context["SSH_USER"] = lambda: self.project.config.ssh_user



import os
from typing import TYPE_CHECKING

from strong_opx.exceptions import ImproperlyConfiguredError
from strong_opx.hcl.runner import HCLRunner

if TYPE_CHECKING:
    from strong_opx.project import Environment


class TerraformRunner(HCLRunner):
    extension = ".tf"
    env_var_prefix = "TF_VAR"

    def get_executable(self) -> str:
        return self.environment.project.config.terraform_executable


def run_terraform(environment: "Environment", command: str, *additional_args: str):
    project = environment.project

    backend_files = list(
        os.path.join(environment.path, file) for file in os.listdir(environment.path) if file.endswith(".tfbackend")
    )

    if len(backend_files) == 0:
        raise ImproperlyConfiguredError(
            "No backend configuration file found. Please create a .tfbackend file in the environment directory."
        )
    elif len(backend_files) > 1:
        raise ImproperlyConfiguredError(
            "Multiple backend configuration files found. Please ensure there is only one "
            ".tfbackend file in the environment directory."
        )

    environ = dict(os.environ)
    terraform_dir = os.path.join(project.path, "terraform")

    # By default, Terraform creates the .terraform directory in the same directory
    # as the other .tf files. Using TF_DATA_DIR lets us create the .terraform dir
    # wherever we want.
    # Source: https://www.terraform.io/cli/config/environment-variables#tf_data_dir
    terraform_data_dir = os.path.join(environment.path, ".terraform")
    environ["TF_DATA_DIR"] = terraform_data_dir

    if command == "init":
        additional_args += (f"-backend-config={backend_files[0]}",)

    # check to see if .terraform has been created
    elif not os.path.exists(terraform_data_dir):
        raise ImproperlyConfiguredError(
            "Please run `strong-opx terraform init` prior to running other Terraform commands."
        )

    runner = TerraformRunner(environment=environment, directory=terraform_dir)
    runner.run(command, env=environ, additional_args=additional_args)



import argparse
from typing import Any

from strong_opx.hcl import run_terraform
from strong_opx.management.command import ProjectCommand
from strong_opx.project import Environment


class Command(ProjectCommand):
    help_text = "Execute Terraform command for specified environment"
    allow_additional_args = True
    parse_known_args = True

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("command", help="Terraform command to execute. e.g. init, plan")
        super().add_arguments(parser)

    def handle(self, environment: Environment, command: str, **options: Any):
        run_terraform(environment, command, *options["additional_args"])



import logging
import random
import string
from functools import lru_cache
from typing import TYPE_CHECKING, Any

from pydantic_core import CoreSchema, core_schema

from strong_opx.providers.discovery import current_provider_name
from strong_opx.utils.module_loading import import_module_attr_if_exists

if TYPE_CHECKING:
    from strong_opx.project import Environment

logger = logging.getLogger(__name__)


class SecretProvider:
    secret_length: int

    def generate_secret(self):
        return "".join(random.choice(string.ascii_letters + string.digits) for _ in range(self.secret_length))

    def get_secret(self, environment: "Environment") -> str:
        raise NotImplementedError()

    @classmethod
    def validate_provider_name(cls, value: str) -> str:
        allowed_secret_providers = tuple(current_secret_providers())

        if value not in allowed_secret_providers:
            raise ValueError(
                f'Unknown secret provider: "{value}". Available provider(s): {", ".join(allowed_secret_providers)}'
            )

        return value

    @classmethod
    def validate_provider_class(cls, d: dict[str, Any]):
        d = d.copy()
        provider = d.pop("provider")
        return current_secret_providers()[provider](**d)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler) -> CoreSchema:
        """
        If the source_type is "SecretProvider", return a schema with the following assertions:
        * The input is a dict, with at least a "provider" key (additional keys are allowed and preserved)
        * The dict is validated by the "validate_provider_class" class method
        * The "provider" key is a string
        * The "provider" value is validated by the "validate_provider_name" class method

        Otherwise, return the result of the handler function.

        Pydantic docs for __get_pydantic_core_schema__():
        https://docs.pydantic.dev/latest/concepts/types/#customizing-validation-with-__get_pydantic_core_schema__

        Param and return comments copied from the Pydantic source code:
        https://github.com/pydantic/pydantic/blob/c067c1b9d0e5ffdf5b0fd53bf0e138cb5924f750/pydantic/main.py#L557-L569

        @param source_type: The class we are generating a schema for. This will generally be the same as the `cls`
                            argument.
        @param handler: Call into Pydantic's internal JSON schema generation. A callable that calls into Pydantic's
                        internal CoreSchema generation logic.
        @return: A `pydantic-core` `CoreSchema`.
        """
        if source_type == SecretProvider:
            # We can't use Literal here because that doesn't work well with OpxString
            provider_schema = core_schema.chain_schema(
                [
                    core_schema.str_schema(),
                    core_schema.no_info_plain_validator_function(cls.validate_provider_name),
                ]
            )

            return core_schema.chain_schema(
                [
                    core_schema.typed_dict_schema(
                        {
                            "provider": core_schema.typed_dict_field(provider_schema, required=True),
                        },
                        extra_behavior="allow",
                    ),
                    core_schema.no_info_plain_validator_function(cls.validate_provider_class),
                ]
            )

        return handler(source_type)


@lru_cache
def current_secret_providers() -> dict[str, SecretProvider]:
    secret_providers = import_module_attr_if_exists(
        f"strong_opx.providers.{current_provider_name()}.secret_provider",
        "SECRET_PROVIDERS",
    )

    return secret_providers or {}



from dataclasses import asdict
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Any, Callable, Collection, get_type_hints

from pydantic.dataclasses import is_pydantic_dataclass
from pydantic_core import CoreSchema, core_schema

from strong_opx.exceptions import ConfigurationError, ErrorDetail
from strong_opx.providers.compute import ComputeInstanceDescription

if TYPE_CHECKING:
    from strong_opx.project import Environment, Project
    from strong_opx.template import Context


class Provider:
    name: str  # Dynamic attribute set by the provider module
    config = None

    compute_instance_id_re: str = None

    def __init__(self, config=None) -> None:
        self.config = config

    def init_project(self, project: "Project") -> None:
        raise NotImplementedError()

    def init_environment(self, environment: "Environment") -> None:
        raise NotImplementedError()

    def describe_compute_instance(self, instance_id: str) -> ComputeInstanceDescription:
        raise NotImplementedError()

    def query_compute_instances(self, ip_address: IPv4Address) -> list[ComputeInstanceDescription]:
        raise NotImplementedError()

    def start_compute_instance(self, instance_ids: Collection[str], wait: bool = True) -> None:
        raise NotImplementedError()

    def stop_compute_instance(self, instance_ids: Collection[str], wait: bool = True) -> None:
        raise NotImplementedError()

    def assume_service_role(self, role: str):
        raise NotImplementedError()

    def update_kubeconfig(self, cluster_name: str, kubeconfig_path: str) -> None:
        raise NotImplementedError()

    def get_additional_context_hooks(self) -> tuple[Callable[["Context"], None], ...]:
        return ()

    def handle_error(self, ex: Exception) -> None:
        raise ex

    def update_config(self, provider_config: dict[str, Any]) -> None:
        config_cls = get_type_hints(type(self)).get("config")
        if config_cls is None:
            if provider_config:
                raise ConfigurationError(
                    ErrorDetail(
                        f"Provider '{self.name}' does not accept any configuration",
                        hint="Remove the provider configuration from your environment file.",
                    )
                )

            return

        config = config_cls(**provider_config)
        for k, v in asdict(config).items():
            if v is None:
                continue

            setattr(self.config, k, v)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler) -> CoreSchema:
        config_cls = get_type_hints(cls).get("config")
        if config_cls is not None:
            if not is_pydantic_dataclass(config_cls):
                raise TypeError(f"{cls.__name__}.config ({repr(config_cls)}) must be a pydantic dataclass")

            schemas = [handler.generate_schema(config_cls), core_schema.no_info_plain_validator_function(source_type)]
        else:
            schemas = [core_schema.no_info_plain_validator_function(lambda x: cls())]

        return core_schema.chain_schema(schemas)



import ast
import os
import sys
import traceback
from typing import TYPE_CHECKING, Any, Callable, Optional

from strong_opx.exceptions import CommandError, ConfigurationError, TemplateError, UndefinedVariableError, VariableError
from strong_opx.template.compiler import CONTEXT_VAR_NAME, INCLUDE_VAR_NAME, OUTPUT_VAR_NAME, TemplateCompiler
from strong_opx.template.lexer import LexerError, TemplateLexer, Token
from strong_opx.template.registry import TEMPLATE_FILTERS
from strong_opx.template.variable import VariableStore
from strong_opx.utils.tracking import Position, get_position

if TYPE_CHECKING:
    from strong_opx.template import Context


class Template:
    module: ast.Module
    variables: VariableStore

    def __init__(self, value: str):
        self.value = value
        self.file_path = get_position(value)[0]

        self.compile()

    def render(self, context: "Context") -> str:
        filename = self.file_path or "<template>"
        local_context: dict[str, Any] = {
            "lines": [],
            INCLUDE_VAR_NAME: self.include,
            CONTEXT_VAR_NAME: context,
        }

        try:
            exec(
                compile(self.module, filename, "exec"),
                self.variables.globals,
                local_context,
            )
        except Exception as e:
            handled_e = self.handle_exception(filename, e)
            if handled_e is None:
                raise

            raise handled_e from e

        if OUTPUT_VAR_NAME in local_context:
            return local_context[OUTPUT_VAR_NAME]

        return "".join(local_context["lines"])

    def handle_exception(self, filename: str, e: Exception) -> Optional[CommandError]:
        """
        Handle an exception raised during template rendering. If the exception is a known template error, it is
        handled and returned. Otherwise, a generic TemplateError is returned.

        :param filename: The filename of the template
        :param e: The exception raised
        :return: The handled exception or None if the exception is not a template error
        """

        for frame, _ in traceback.walk_tb(sys.exc_info()[-1]):
            if frame.f_code.co_filename == filename:
                break
        else:
            return None

        if isinstance(e, VariableError):
            e.errors[0].file_path = self.file_path

            ref = self.variables.get_ref(e.var_name)
            if ref is not None:
                e.errors[0].start_pos = ref.start_pos
                e.errors[0].end_pos = ref.end_pos

                return e

        if isinstance(e, UndefinedVariableError):
            for i, name in enumerate(e.names):
                e.errors[i].file_path = self.file_path

                ref = self.variables.get_ref(name)
                if ref is not None:
                    e.errors[i].start_pos = ref.start_pos
                    e.errors[i].end_pos = ref.end_pos

            return e

        if isinstance(e, ConfigurationError):
            for error in e.errors:
                error.file_path = self.file_path
                error.start_pos = Position(line=frame.f_lineno, column=None)

            return e

        try:
            from strong_opx.providers import current_provider_error_handler

            current_provider_error_handler(e)
        except CommandError as ex:
            return ex
        except Exception as ex:
            e = ex  # Fall back to generic error handling

        return TemplateError(
            f"({e.__class__.__name__}) {e}",
            file_name=get_position(self.value)[0],
            start_pos=Position(line=frame.f_lineno, column=None),
        )

    def compile(self):
        try:
            tokens = TemplateLexer(self.value).tokenize()
        except LexerError as e:
            raise TemplateCompiler(self.value).syntax_error(e.message, e.start_pos, e.end_pos)

        # If there are multiple tokens, render as string otherwise keep the original datatype
        compiler = TemplateCompiler(self.value, as_string=len(tokens) > 1)
        ops_stack = []

        for token in tokens:
            if isinstance(token, Token):
                compiler.compile_constant(token.value, token.position)
            else:
                start, content, end = token

                if start.value == "{{":  # An expression to evaluate.
                    compiler.compile_expression(content.value, content.position)

                elif start.value.startswith("${"):  # $-style expression.
                    expr = start.value + content.value + end.value
                    compiler.compile_legacy_expression(expr, start.position)

                elif start.value == "{%":  # Action tag: split into words and parse further.
                    action_tag, tag_args = compiler.split_on_whitespace(content.value)

                    if action_tag == "if":  # An if statement: evaluate the expression to determine if.
                        ops_stack.append(("if", content.position, content.position + 2))
                        compiler.compile_if(tag_args, content.position)
                    elif action_tag == "for":  # A loop: iterate over expression result.
                        ops_stack.append(("for", content.position, content.position + 3))
                        compiler.compile_for(tag_args, content.position)
                    elif action_tag == "raw":
                        # When no-args are given, that scenario is handled separately
                        if tag_args:
                            raise compiler.syntax_error(
                                "raw action tag does not take any argument", content.position, content.end_position
                            )
                    elif action_tag == "include":
                        compiler.compile_include(tag_args, start.position, content.position, content.end_position)

                    elif action_tag.startswith("end"):  # End-something. Pop the ops stack.
                        if tag_args:
                            raise compiler.syntax_error(
                                "End block does not take any argument", content.position, content.end_position
                            )

                        end_what = action_tag[3:]
                        if not ops_stack:
                            raise compiler.syntax_error("Unexpected end block", start.position, start.end_position)

                        start_what = ops_stack.pop()[0]
                        if start_what != end_what:
                            raise compiler.syntax_error(
                                f"Expecting end{start_what} block, got end{end_what}",
                                start.position,
                                start.end_position,
                            )

                        compiler.close_block()
                    elif action_tag == "else":
                        if tag_args:
                            raise compiler.syntax_error(
                                "else block does not take any argument", content.position, content.end_position
                            )

                        if not ops_stack or ops_stack[-1][0] != "if":
                            raise compiler.syntax_error("Unexpected else block", content.position, content.end_position)

                        compiler.start_else_block(content.position, content.end_position)
                    else:
                        raise compiler.syntax_error(
                            f"Unknown action tag: {action_tag}", content.position, content.end_position
                        )

                elif start.value == "{% raw %}":
                    compiler.compile_constant(content.value, content.position)

                else:
                    raise RuntimeError(f"Internal error: unknown token type: {start.value} at {start.position}")

        if ops_stack:
            value, start_pos, end_pos = ops_stack.pop()
            raise compiler.syntax_error(f"Unclosed tag: {value}", start_pos, end_pos)

        self.module = compiler.finalize()
        self.variables = compiler.variables

    def include(self, template_name: str, *, context: "Context", indent=0) -> str:
        template_dir = "."
        if self.file_path:
            template_dir = os.path.dirname(self.file_path)

        template_path = os.path.join(template_dir, template_name)
        if not os.path.isfile(template_path):
            raise FileNotFoundError(f"Included template '{template_name}' not found in '{template_dir}'")

        with open(template_path, "r") as f:
            content = f.read()

        rendered_content = Template(content).render(context)
        if indent > 0:
            rendered_content = rendered_content.replace("\n", "\n" + (" " * indent))

        return rendered_content

    @classmethod
    def register_filter(cls, name: str, func: Callable = None):
        if name in TEMPLATE_FILTERS:
            raise ValueError(f"Filter {name} is already exists")

        def register(f: Callable):
            TEMPLATE_FILTERS[name] = f
            return f

        if func is None:
            return register

        return register(func)



import os
import re
from datetime import datetime
from functools import lru_cache
from typing import Any

import boto3

from strong_opx.config import system_config
from strong_opx.providers.aws.credentials import AWSCredential, aws_credentials

SERVICE_ROLE_ARN_RE = re.compile(r"^arn:aws:iam::[0-9]+:role/.+$")

DUPLICATE_HYPHENS_RE = re.compile(r"-{2,}")
ROLE_SESSION_NAME_INVALID_CHARS_RE = re.compile(r"[^\w+=,.@-]")


@lru_cache(maxsize=1)
def get_caller_identity() -> dict[str, Any]:
    return boto3.client("sts").get_caller_identity()


def get_current_account_id() -> str:
    profile = os.getenv("AWS_PROFILE", "default")
    account_id = system_config.get(f"aws:{profile}", "account_id", fallback=None)
    if account_id is None:
        account_id = get_caller_identity()["Account"]
        system_config.set(f"aws:{profile}", "account_id", account_id)

    return account_id


def get_current_user_id() -> str:
    return get_caller_identity()["UserId"]


def _assume_role(arn: str) -> tuple[AWSCredential, datetime]:
    client = boto3.client("sts")

    role_session_name = DUPLICATE_HYPHENS_RE.sub(
        "-", ROLE_SESSION_NAME_INVALID_CHARS_RE.sub("-", f"StrongOpX-{get_current_user_id()}")
    )

    response = client.assume_role(
        RoleArn=arn,
        RoleSessionName=role_session_name,
    )

    credentials_dict = response["Credentials"]
    credential = AWSCredential(
        aws_access_key_id=credentials_dict["AccessKeyId"],
        aws_secret_access_key=credentials_dict["SecretAccessKey"],
        aws_session_token=credentials_dict["SessionToken"],
    )
    return credential, credentials_dict["Expiration"]


def assume_role(arn: str) -> AWSCredential:
    if not SERVICE_ROLE_ARN_RE.match(arn):
        raise ValueError(f"Invalid service role ARN: {arn}")

    credential = aws_credentials.get_credential(arn)
    if credential is None:
        print(f"Assuming role: {arn}")
        credential, expires = _assume_role(arn)
        aws_credentials.set_credential(arn, credential, expires)
    else:
        print(f"Using cached credentials for: {arn}")

    return credential



import os
import subprocess
import sys
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

import tabulate
from colorama import Style

from strong_opx import yaml
from strong_opx.exceptions import YAMLError
from strong_opx.platforms.deployments.base import ConfigDeploymentProvider, NodeConfig

if TYPE_CHECKING:
    from strong_opx.platforms import KubernetesPlatform


@dataclass
class DeploymentStatus:
    resource_type: str
    resource_name: str
    status: str

    __slots__ = ("resource_type", "resource_name", "status")

    @classmethod
    def from_kubectl_output(cls, output) -> "DeploymentStatus":
        resource, status = output.split(" ")
        type_, name = resource.split("/")
        return cls(
            resource_type=type_.split(".")[0].lower(),
            resource_name=name,
            status=status,
        )


class DeploymentSummary:
    def __init__(self):
        self.lines: list[DeploymentStatus] = []

    def print(self):
        print(
            tabulate.tabulate(
                [(line.resource_type, line.resource_name, line.status) for line in self.lines],
                headers=["Resource", "Name", "Status"],
            )
        )

    @cached_property
    def unchanged_deployments(self) -> list[str]:
        return [
            line.resource_name
            for line in self.lines
            if line.resource_type == "deployment" and line.status == "unchanged"
        ]

    @classmethod
    def parse(cls, kubectl_output: str) -> "DeploymentSummary":
        summary = cls()
        summary.lines = [DeploymentStatus.from_kubectl_output(line) for line in kubectl_output.splitlines()]

        return summary


def preprocess_config_file(node: NodeConfig, manifest_directory: str, manifest_file: str) -> list[dict]:
    full_path = os.path.join(manifest_directory, manifest_file)

    try:
        documents = list(yaml.load_all(full_path))
    except YAMLError as e:
        if not sys.stdin.isatty():
            raise

        print(
            "\nThere is an error while loading rendered Kubernetes manifest. Actual Kubernetes manifest is:\n",
            f"  {Style.BRIGHT}{os.path.join(node.path, manifest_file)}{Style.RESET_ALL}",
        )
        print(str(e), file=sys.stderr)

        print(
            f"\nA temporary file has been created for you to view the rendered manifest:\n"
            f"  {Style.BRIGHT}{full_path}{Style.RESET_ALL}\n",
            file=sys.stderr,
        )

        print(
            "Any changes made to temporary file will be lost. Make changes to the original file instead.",
            file=sys.stderr,
        )
        input("Press Enter when you are done viewing the file to continue...")
        sys.exit(255)

    return documents


class KubeCtlDeploymentProvider(ConfigDeploymentProvider):
    name = "kubectl"
    allowed_extensions = (".yml", ".yaml")
    platform: "KubernetesPlatform"
    rollout_status_supported_kinds = ("deployment", "statefulset")

    def deploy(self, node: NodeConfig, manifest_directory: str) -> None:
        rollout_supported_kinds = []

        for manifest_file in os.listdir(manifest_directory):
            documents = preprocess_config_file(node, manifest_directory, manifest_file)

            for document in documents:
                kind = document.get("kind", "").lower()

                if kind in self.rollout_status_supported_kinds:
                    metadata = document.get("metadata") or {}
                    name = metadata.get("name")
                    namespace = metadata.get("namespace", "default")

                    rollout_supported_kinds.append((kind, name, namespace))

        result = self.platform.kubectl("apply", "-f", manifest_directory, stdout=subprocess.PIPE)
        summary = DeploymentSummary.parse(result.stdout.decode("utf8"))
        summary.print()

        for kind, name, namespace in rollout_supported_kinds:
            self.platform.kubectl("rollout", "status", kind, name, "-n", namespace)

        for kind, name, namespace in rollout_supported_kinds:
            if name in summary.unchanged_deployments:
                # Restart deployment as it was not updated
                self.platform.kubectl("rollout", "restart", "deployment", name, "-n", namespace)
                self.platform.kubectl("rollout", "status", "deployment", name, "-n", namespace)



from typing import Any, TextIO, Union

import yaml
from yaml.representer import SafeRepresenter

from strong_opx.utils.tracking import OpxFloat, OpxInteger, OpxMapping, OpxSequence, OpxString


class OpxYAMLDumper(yaml.Dumper):
    def ignore_aliases(self, data):
        # We don't want to use aliases in YAML, because they're not supported by all tools.
        return True


OpxYAMLDumper.add_representer(OpxString, SafeRepresenter.represent_str)
OpxYAMLDumper.add_representer(OpxInteger, SafeRepresenter.represent_int)
OpxYAMLDumper.add_representer(OpxFloat, SafeRepresenter.represent_float)
OpxYAMLDumper.add_representer(OpxSequence, SafeRepresenter.represent_list)
OpxYAMLDumper.add_representer(OpxMapping, SafeRepresenter.represent_dict)


def dump_all(data: list[Any], target: Union[str, TextIO]) -> None:
    if isinstance(target, str):
        with open(target, "w") as f:
            return yaml.dump_all(data, f, Dumper=OpxYAMLDumper)
    else:
        return yaml.dump_all(data, target, Dumper=OpxYAMLDumper)


def dump(data: Any, target: Union[str, TextIO]) -> None:
    dump_all([data], target)



import re
from typing import Any, Optional

import boto3
from colorama import Fore as TextColor

from strong_opx.exceptions import CommandError
from strong_opx.management.command import BaseCommand
from strong_opx.providers.aws import AWSCredentialConfig
from strong_opx.providers.aws.credentials import mfa_profile_name
from strong_opx.utils.prompt import select_prompt

MFA_ARN_RE = re.compile(r"arn:aws:iam::(?P<account_id>\d+):mfa/(?P<username>.+)")
USER_ARN_RE = re.compile(r"arn:aws:iam::(?P<account_id>\d+):user/(?P<username>.+)")


def prompt_until_valid(prompt: str) -> str:
    while True:
        value = input(prompt).strip()
        if value:
            return value


def select_mfa_device(session) -> Optional[str]:
    caller_arn = session.client("sts").get_caller_identity()["Arn"]
    match = USER_ARN_RE.match(caller_arn)
    if not match:
        return None  # Not an IAM user; assume no MFA

    username = match.group("username")
    all_2fa_devices = session.client("iam").list_mfa_devices(UserName=username)["MFADevices"]

    mfa_devices = [device for device in all_2fa_devices if MFA_ARN_RE.match(device["SerialNumber"])]

    if not mfa_devices:
        return None  # No MFA device

    if len(mfa_devices) > 1:
        mfa_serial_number = select_prompt(
            "Select MFA Device",
            [device["SerialNumber"] for device in mfa_devices],
        )
    else:
        mfa_serial_number = mfa_devices[0]["SerialNumber"]

    return mfa_serial_number


class Command(BaseCommand):
    help_text = "Configure AWS credentials"

    def handle(self, **options: Any) -> None:
        profile_name = prompt_until_valid("Profile name: ")

        aws_config = AWSCredentialConfig()
        aws_access_key_id = prompt_until_valid("AWS Access Key ID: ")
        aws_secret_access_key = prompt_until_valid("AWS Secret Access Key: ")

        boto3_session = boto3.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )

        section_name = profile_name
        mfa_device = select_mfa_device(boto3_session)
        if mfa_device:
            section_name = mfa_profile_name(profile_name)

        if aws_config.has_section(section_name):
            raise CommandError(f'Profile "{profile_name}" already exists in {aws_config.config_path}')

        aws_config.add_section(section_name)
        aws_config.set(section_name, "aws_access_key_id", aws_access_key_id)
        aws_config.set(section_name, "aws_secret_access_key", aws_secret_access_key)

        if mfa_device:
            aws_config.set(section_name, "aws_mfa_device", mfa_device)

        aws_config.save()
        print(f'{TextColor.GREEN}Profile "{profile_name}" configured successfully{TextColor.RESET}')

        if mfa_device:
            print(
                "\n"
                "MFA is enabled for this profile. Run following command to get temporary credentials:\n\n"
                f"    strong-opx aws:mfa --profile {profile_name} <mfa-token-code>\n"
            )



import logging
import os
import sys
from functools import cached_property
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Callable, Collection

import boto3
from botocore.exceptions import ClientError, NoCredentialsError

from strong_opx.project import Environment, Project
from strong_opx.providers.aws.compute import describe_instances
from strong_opx.providers.aws.config import AWSConfig, get_aws_config
from strong_opx.providers.aws.context_hooks import import_and_clean_environ_hook, update_environ_hook
from strong_opx.providers.aws.errors import handle_boto_error
from strong_opx.providers.aws.iam import assume_role
from strong_opx.providers.compute import ComputeInstanceDescription
from strong_opx.providers.provider import Provider
from strong_opx.utils.shell import shell

if TYPE_CHECKING:
    from strong_opx.template import Context

logger = logging.getLogger(__name__)


class AWSProvider(Provider):
    config: AWSConfig
    compute_instance_id_re = r"^i-[0-9a-f]+$"

    def get_additional_context_hooks(self) -> tuple[Callable[["Context"], None], ...]:
        return (
            self.update_context,
            import_and_clean_environ_hook,
            update_environ_hook,
        )

    def update_context(self, context: "Context"):
        aws_profile = self.default_aws_profile
        if aws_profile is not None:
            context["AWS_PROFILE"] = aws_profile

        context.update(self.config.dict())

    @cached_property
    def default_aws_profile(self):
        project = Project.current()

        return project.config.get("aws", "aws_profile", fallback=None)

    def init_project(self, project: "Project"):
        session_kwargs = {}
        client_kwargs = {
            "region_name": self.config.region if self.config.region else "us-east-1",
        }

        ops_bucket_name = f"{project.name}-ops"
        aws_profile = self.default_aws_profile
        if aws_profile is not None:
            session_kwargs["profile_name"] = aws_profile

        logger.info(f"Creating S3 bucket: {ops_bucket_name}...")

        s3_client = boto3.Session(**session_kwargs).client("s3", **client_kwargs)
        s3_client.create_bucket(
            ACL="private",
            Bucket=ops_bucket_name,
            ObjectOwnership="BucketOwnerEnforced",
        )
        s3_client.put_bucket_tagging(
            Bucket=ops_bucket_name,
            Tagging={
                "TagSet": [
                    {
                        "Key": "Description",
                        "Value": f"Bucket used by Terraform to store the state of the "
                        f"infrastructure of the {project.name} project",
                    },
                ]
            },
        )

        logger.info(f"Block public access to bucket and its objects...")
        s3_client.put_public_access_block(
            Bucket=ops_bucket_name,
            PublicAccessBlockConfiguration={
                "BlockPublicAcls": True,
                "IgnorePublicAcls": True,
                "BlockPublicPolicy": True,
                "RestrictPublicBuckets": True,
            },
        )

        logger.info(f"Enabling Versioning on bucket...")
        s3_client.put_bucket_versioning(Bucket=ops_bucket_name, VersioningConfiguration={"Status": "Enabled"})

        logger.info(f"Enabling Server-Side Encryption on bucket...")
        s3_client.put_bucket_encryption(
            Bucket=ops_bucket_name,
            ServerSideEncryptionConfiguration={
                "Rules": [{"ApplyServerSideEncryptionByDefault": {"SSEAlgorithm": "AES256"}}]
            },
        )

    def init_environment(self, environment: "Environment") -> None:
        dynamodb_client = boto3.client("dynamodb")
        dynamodb_client.create_table(
            AttributeDefinitions=[
                {"AttributeName": "LockID", "AttributeType": "S"},
            ],
            TableName=f"{environment.project.name}-{environment.name}-terraform-state",
            KeySchema=[
                {"AttributeName": "LockID", "KeyType": "HASH"},
            ],
            ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
            Tags=[
                {
                    "Key": "Description",
                    "Value": (
                        "Table used by Terraform to ensure that only one user is modifying the"
                        f"{environment.name} resources of {environment.project.name} at the same time"
                    ),
                },
                {
                    "Key": "Environment",
                    "Value": environment.name,
                },
                {
                    "Key": "Project",
                    "Value": environment.project.name,
                },
            ],
        )

    def query_compute_instances(self, ip_address: IPv4Address) -> list[ComputeInstanceDescription]:
        return describe_instances(
            Filters=[
                {
                    "Name": (
                        "network-interface.addresses.private-ip-address" if ip_address.is_private else "ip-address"
                    ),
                    "Values": [str(ip_address)],
                }
            ]
        )

    def describe_compute_instance(self, instance_id: str) -> ComputeInstanceDescription:
        instances = describe_instances(InstanceIds=[instance_id])
        if len(instances) == 0:
            raise ValueError(f'Unable to find an instance with ID "{instance_id}"')

        return instances[0]

    def start_compute_instance(self, instance_ids: Collection[str], wait: bool = True) -> None:
        client = boto3.client("ec2")
        client.start_instances(InstanceIds=instance_ids)

        if wait:
            waiter = client.get_waiter("instance_status_ok")
            logger.info(f"Waiting for {len(instance_ids)} instance(s) to start...")
            waiter.wait(InstanceIds=instance_ids)

    def stop_compute_instance(self, instance_ids: Collection[str], wait: bool = True) -> None:
        client = boto3.client("ec2")
        client.stop_instances(InstanceIds=instance_ids)

        if wait:
            waiter = client.get_waiter("instance_stopped")
            logger.info(f"Waiting for {len(instance_ids)} instance(s) to stop...")
            waiter.wait(InstanceIds=instance_ids)

    def assume_service_role(self, role: str) -> None:
        os.environ.update(assume_role(role).dict())
        os.environ.pop("AWS_PROFILE", None)  # Remove AWS_PROFILE env if it exists

    def handle_error(self, ex: Exception) -> None:
        if isinstance(ex, NoCredentialsError):
            print("error: Unable to locate AWS Credentials", file=sys.stderr)
            exit(10)

        elif isinstance(ex, ClientError):
            try:
                handle_boto_error(ex)
            except ClientError as e:
                error_info = e.response.get("Error", {})
                error_code = error_info.get("Code")
                error_message = error_info.get("Message")

                print(f"boto3 error: ({error_code}) {error_message or e}", file=sys.stderr)
                exit(11)

        super().handle_error(ex)

    def update_kubeconfig(self, cluster_name: str, kubeconfig_path: str) -> None:
        shell(
            [
                "aws",
                "eks",
                "--region",
                get_aws_config("region"),
                "update-kubeconfig",
                "--name",
                cluster_name,
                "--kubeconfig",
                kubeconfig_path,
            ],
        )



import functools
import re
from typing import List, NamedTuple, Optional, Tuple, Union

RE_RAW_START = re.compile(r"\{%\s*raw\s*%}")
RE_RAW_END = re.compile(r"\{%\s*endraw\s*%}")

RE_COMMENT_START = re.compile(r"\{#")
RE_COMMENT_END = re.compile(r"#}")

RE_VARIABLE_START = re.compile(r"\{\{")
RE_VARIABLE_END = re.compile(r"}}")

RE_BLOCK_START = re.compile(r"\{%")
RE_BLOCK_END = re.compile(r"%}")

RE_LEGACY_VARIABLE_START = re.compile(r"\${{?")
RE_LEGACY_VARIABLE_END = re.compile(r"}?}")


class LexerError(Exception):
    def __init__(self, message: str, start_pos: int, end_pos: Optional[int] = None):
        self.message = message
        self.start_pos = start_pos
        self.end_pos = end_pos


class Token(NamedTuple):
    value: str
    position: int

    @property
    def end_position(self) -> int:
        return self.position + len(self.value)


class TemplateLexer:
    def __init__(self, value: str):
        self.template = value
        self._tokens = []

    def tokenize(self) -> List[Union[Token, Tuple[Token, Token, Token]]]:
        self._tokens = []
        length = len(self.template)
        position = 0

        pattern_handlers = (
            # Check for {% raw %} first
            (RE_RAW_START, self._handle_raw_block),
            (RE_COMMENT_START, self._skip_comment),  # Check for comments
            (RE_VARIABLE_START, functools.partial(self._handle_tag, RE_VARIABLE_END)),
            (RE_BLOCK_START, functools.partial(self._handle_tag, RE_BLOCK_END)),
            (
                RE_LEGACY_VARIABLE_START,
                functools.partial(self._handle_tag, RE_LEGACY_VARIABLE_END, strip_whitespaces=False),
            ),
        )

        while position < length:
            text_start = position
            next_delim_pos = length

            next_delim_match = None
            next_delim_handler = None

            for pattern, handler in pattern_handlers:
                match = pattern.search(self.template, position)
                if match and match.start() < next_delim_pos:
                    next_delim_pos = match.start()

                    next_delim_match = match
                    next_delim_handler = handler

            if next_delim_pos > text_start:
                text = self.template[text_start:next_delim_pos]
                self._tokens.append(Token(text, text_start))

            if next_delim_handler:
                position = next_delim_handler(next_delim_match.end(), next_delim_match)
            else:
                position = next_delim_pos

        return self._tokens

    def _skip_comment(self, content_start: int, start_match: re.Match) -> int:
        end_match = RE_COMMENT_END.search(self.template, content_start)
        if end_match:
            return end_match.end()

        raise LexerError("Unclosed tag", start_match.start(), start_match.end())

    def _handle_raw_block(self, content_start: int, start_match: re.Match) -> int:
        end_match = RE_RAW_END.search(self.template, content_start)
        if end_match is None:
            raise LexerError("Unclosed tag", start_match.start(), start_match.end())

        content_end = end_match.start()
        if content_end > content_start:
            content = self.template[content_start:content_end]
            self._tokens.append(
                (
                    Token("{% raw %}", start_match.start()),
                    Token(content, content_start),
                    Token("{% endraw %}", end_match.start()),
                )
            )

        return end_match.end()

    def _handle_tag(self, end_re: re.Pattern, content_start: int, start_match: re.Match, strip_whitespaces=True) -> int:
        end_match = end_re.search(self.template, content_start)
        if end_match is None:
            raise LexerError("Unclosed tag", start_match.start(), start_match.end())

        content_end = end_match.start()
        if content_end > content_start:
            content = self.template[content_start:content_end]

            if strip_whitespaces:
                actual_length = len(content)
                content = content.lstrip()
                content_start += actual_length - len(content)

                content = content.rstrip()

            self._tokens.append(
                (
                    Token(start_match.group(0), start_match.start()),
                    Token(content, content_start),
                    Token(end_match.group(0), end_match.start()),
                )
            )

        return end_match.end()



import os
import subprocess
from typing import TYPE_CHECKING

from strong_opx.exceptions import UndefinedVariableError
from strong_opx.hcl.extractor import HCLVariableExtractor
from strong_opx.utils.shell import shell

if TYPE_CHECKING:
    from strong_opx.project import Environment


class HCLRunner:
    extension: str
    env_var_prefix: str

    def __init__(self, environment: "Environment", directory: str):
        self.environment = environment
        self.directory = directory

    def extract_vars(self) -> dict[str, str]:
        env_dict = {}
        extractor = HCLVariableExtractor()

        context = self.environment.context

        for filename in os.listdir(self.directory):
            if filename.endswith(self.extension):
                file_path = os.path.join(self.directory, filename)
                with open(file_path) as f:
                    extractor.extract(file_path, f)

        missing_vars = []
        for var in extractor.required_vars:
            if var not in context:
                missing_vars.append(var)
                continue

            env_dict[f"{self.env_var_prefix}_{var}"] = str(context[var])

        if missing_vars:
            raise UndefinedVariableError(*missing_vars)

        for var in extractor.optional_vars:
            if var in context:
                env_dict[f"{self.env_var_prefix}_{var}"] = str(context[var])

        return env_dict

    def get_executable(self) -> str:
        raise NotImplementedError()

    def _run(self, args: tuple[str, ...], env: dict[str, str], **kwargs) -> subprocess.CompletedProcess:
        return shell(args, env=env, cwd=self.directory, **kwargs)

    def run(
        self, command: str, env: dict[str, str] = None, additional_args: tuple[str, ...] = ()
    ) -> subprocess.CompletedProcess:
        if env is None:
            env = dict(os.environ)

        env.update(self.extract_vars())
        args = (self.get_executable(), command, *additional_args)
        return self._run(args, env=env)



import os

from strong_opx.config.base import Config, SystemConfig
from strong_opx.config.hierarchical import HierarchicalConfig
from strong_opx.config.opx import DEFAULT_TEMPLATING_ENGINE, StrongOpxConfig

PROJECT_CONFIG_FILE = "strong-opx.yml"

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(os.path.realpath(__file__))))
TEMPLATES_DIR = os.path.join(BASE_DIR, "templates")
ANSIBLE_FILTER_PLUGINS = os.path.join(BASE_DIR, "ansible", "filter_plugins")

CONFIG_DIR = os.path.join(os.path.expanduser("~"), ".strong-opx")
CACHE_DIR = os.path.join(CONFIG_DIR, "cache")

system_config = SystemConfig(os.path.join(CONFIG_DIR, "config"))
opx_config = StrongOpxConfig()



import os

from botocore.exceptions import ClientError
from colorama import Style

from strong_opx.exceptions import CommandError, RepositoryNotFoundException
from strong_opx.providers.aws.credentials import AWSCredentialConfig


class CredentialExpiredError(CommandError):
    def __str__(self):
        profile_name = os.getenv("AWS_PROFILE", "default")

        aws_config = AWSCredentialConfig()
        if not aws_config.are_credential_expired(profile_name):
            return super().__str__()

        return (
            f'Credentials for profile "{profile_name}" have expired'
            f"\n\n"
            "MFA is enabled for this profile. Run following command to get temporary credentials:"
            "\n\n"
            f"{Style.BRIGHT}    strong-opx aws:mfa --profile {profile_name} <mfa-token-code>{Style.RESET_ALL}\n"
        )


AWS_ERROR_CODE_TO_EXCEPTION = {
    "ExpiredToken": CredentialExpiredError,
    "RequestExpired": CredentialExpiredError,
    "ExpiredTokenException": CredentialExpiredError,
    "RepositoryNotFoundException": RepositoryNotFoundException,
}


def handle_boto_error(ex: ClientError, ignore: tuple[str, ...] = ()) -> None:
    error = (ex.response or {}).get("Error", {})
    code = error.get("Code")
    if code in ignore:
        return  # Ignore this exception

    constructor = AWS_ERROR_CODE_TO_EXCEPTION.get(code)
    if not constructor:
        raise

    message = error.get("Message", str(ex))
    raise constructor(message)



import weakref
from collections import ChainMap
from typing import Any

from strong_opx import yaml
from strong_opx.exceptions import UndefinedVariableError, VariableError
from strong_opx.template.object_template import ObjectTemplate
from strong_opx.utils.mapping import LazyDict


class Context(LazyDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for k, v in self._data.items():
            if callable(v):
                self.set_lazy(k, v)

        self._children: list[weakref.ref] = []

    def __missing__(self, key):
        raise UndefinedVariableError(key)

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self:
            # If the key already exists in the current context, we need to update all
            # children contexts to have old value. Otherwise, they will get the new value
            # from the parent context

            current_value = self._data[key]
            i = len(self._children) - 1

            while i >= 0:
                child_ref = self._children[i]
                child = child_ref()
                if child is None:
                    # Child has been garbage collected. Remove it from the list
                    self._children.pop(i)
                else:
                    child._data[key] = current_value

                i -= 1

        if callable(value):
            self.set_lazy(key, value)
        else:
            super().__setitem__(key, value)

    def __delitem__(self, key) -> None:
        """
        Deleting variables is not supported for two reasons:
        1. As of now, there is no known use case for deleting variables from a Context
        2. A Context may use a ChainMap to store data, and deleting a variable from a ChainMap would require some
           extra thought on how we would want to handle that. Should we delete the variable from "parent" and/or "child"
           chains? Should we just set the value to None in the first map of the ChainMap? If we set to None, then how
           should the 'in' operator behave; should "deleted_key in context" return False or True? There are a lot of
           questions that need to be answered before we can implement this feature (a feature that is not needed now).

        :param key: Key to delete
        :return: (None)
        :raise: Raises a NotImplementedError
        """
        raise NotImplementedError("Deleting variables is not supported")

    @property
    def initial_vars(self) -> set[str]:
        """
        Returns the set of "initial variables" in the current Context. Initial variables are the system variables
        that are defined in the first Context, and are not inherited/chained from any other Context.

        This method returns an empty set when this Context does not use a ChainMap for self._data. The Context used
        to represent the system variables is expected to use a dict for self._data, and will therefore return an empty
        set from this method.

        :return: Set of initial variables
        """
        if isinstance(self._data, ChainMap):
            return set(self._data.maps[-1])

        return set()

    def chain(self) -> "Context":
        context = type(self)()
        self._children.append(weakref.ref(context))

        if isinstance(self._data, ChainMap):
            context._data = self._data.new_child(context._data)
        else:
            context._data = ChainMap(context._data, self._data)
        return context

    def as_dict(self, exclude_initial: bool = False) -> dict[str, Any]:
        if exclude_initial:
            initial_vars = self.initial_vars
            return {k: self[k] for k in self if k not in initial_vars}

        return dict(self)

    def require(self, *names) -> dict[str, Any]:
        unknowns = []
        resolved = {}

        for n in names:
            if n in self:
                resolved[n] = self[n]
            else:
                unknowns.append(n)

        if unknowns:
            raise UndefinedVariableError(*unknowns)

        return resolved

    def load_from_file(self, file_path: str) -> None:
        vars_ = yaml.load(file_path)
        if vars_:
            vars_ = ObjectTemplate(self).render(vars_)
            initial_vars = self.initial_vars
            for k, v in vars_.items():
                if k in initial_vars:
                    raise VariableError(f"{k} is protected variable and cannot be altered.", var_name=k)

                self[k] = v



import itertools
import logging
import os
import tempfile
from contextlib import contextmanager
from typing import TYPE_CHECKING, Annotated, Generator, Optional

from pydantic import AfterValidator, BaseModel, model_validator

from strong_opx import config, yaml
from strong_opx.exceptions import CommandError, ComputeInstanceError
from strong_opx.platforms import Platform
from strong_opx.providers import ComputeInstance
from strong_opx.providers.compute import ComputeInstanceState
from strong_opx.utils.shell import shell

if TYPE_CHECKING:
    from strong_opx.template import Context

logger = logging.getLogger(__name__)
SUPPORTED_SSH_METHODS = ("direct", "bastion", "aws_ssm")


def validate_ssh_method(value: str) -> str:
    if value not in SUPPORTED_SSH_METHODS:
        raise ValueError(f"Unsupported SSH method: {value}")

    return value


class GenericPlatformConfig(BaseModel):
    ssh_method: Annotated[str, AfterValidator(validate_ssh_method)] = None
    hosts: dict[str, list[ComputeInstance]]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.ssh_method is None:
            self.ssh_method = "bastion" if "bastion" in self.hosts else "direct"

    @model_validator(mode="after")
    def check_bastion_config(self) -> "GenericPlatformConfig":
        if self.ssh_method == "bastion" and "bastion" not in self.hosts:
            raise ValueError('SSH method is set to "bastion" but no bastion host is configured')

        if "bastion" in self.hosts and len(self.hosts["bastion"]) > 1:
            raise ValueError("Only one bastion host is allowed")

        for group, group_hosts in self.hosts.items():
            if group == "bastion":
                group_hosts[0].hostname = group
                continue

            for host_index, host in enumerate(group_hosts):
                host.hostname = f"{group}:{host_index}"

        return self


class GenericPlatform(Platform):
    config_class = GenericPlatformConfig

    if TYPE_CHECKING:
        ssh_method: str
        hosts: dict[str, list[ComputeInstance]]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for host_groups in self.hosts.values():
            for host in host_groups:
                host.platform = self

    def init_context(self, context: "Context"):
        def read_git_ssh_key():
            with open(self.project.config.git_ssh_key) as f:
                return f.read()

        context["GIT_IDENTITY"] = read_git_ssh_key

    def deploy(self) -> None:
        playbook_path = os.path.join(self.project.path, "playbooks", "deploy.yml")
        if not os.path.exists(playbook_path):
            raise CommandError(f"Missing playbooks/deploy.yml")

        with self.ensure_instances_are_running():
            self.ansible_playbook(playbook_path)

    def ansible_playbook(
        self,
        playbook: str,
        host_groups: list[str] = None,
        additional_args: tuple[str, ...] = None,
    ):
        with self.ansible_host_inventory(host_groups) as inventory, self.ansible_extra_vars() as extras:
            args = (
                self.project.config.ansible_playbook_executable,
                playbook,
                "-i",
                inventory,
                "-u",
                self.project.config.ssh_user,
                "--private-key",
                self.project.config.ssh_key,
                "--extra-vars",
                f"@{extras}",
            )

            os.environ["STRONG_OPX_PROJECT"] = self.project.name
            os.environ["STRONG_OPX_ENVIRONMENT"] = self.environment.name
            os.environ["ANSIBLE_HOST_KEY_CHECKING"] = "False"
            os.environ["ANSIBLE_FILTER_PLUGINS"] = config.ANSIBLE_FILTER_PLUGINS
            os.environ["ANSIBLE_PIPELINING"] = "True"
            os.environ["ANSIBLE_SSH_PIPELINING"] = "True"
            if additional_args:
                args += additional_args

            shell(args, cwd=self.project.path)

    def ssh_proxy_command(self, hostname: str = None) -> Optional[str]:
        if self.ssh_method == "direct":
            return None

        if self.ssh_method == "bastion":
            if hostname == "bastion":
                return None

            bastion = self.hosts["bastion"][0]
            return (
                f"ssh -A -o StrictHostKeyChecking=no -i {self.project.config.ssh_key} "
                f"-W %h:%p {self.project.config.ssh_user}@{bastion.ip_address}"
            )

        return (
            f'sh -c "aws ssm start-session --target %h --document-name '
            f'AWS-StartSSHSession --parameters portNumber=%p"'
        )

    def resolve_instance(self, hostname: str) -> list[ComputeInstance]:
        if hostname == "all":
            return list(itertools.chain.from_iterable(self.hosts.values()))

        if ":" in hostname:
            for instance in itertools.chain.from_iterable(self.hosts.values()):
                if instance.hostname == hostname:
                    return [instance]

            raise ComputeInstanceError(f'Compite Instance "{hostname}" not found')

        if hostname in self.hosts:
            group_hosts = self.hosts[hostname]
            if not group_hosts:
                raise ValueError(f'Host group "{hostname}" has no hosts')

            return group_hosts

        try:
            instance = ComputeInstance.parse(hostname)
        except ValueError:
            raise ComputeInstanceError(f"Unknown host group or invalid format: {hostname}")

        instance.platform = self
        return [instance]

    def get_ssh_host(self, instance: ComputeInstance) -> str:
        if self.ssh_method == "aws_ssm":
            return instance.instance_id

        return str(instance.ip_address)

    @contextmanager
    def ensure_instances_are_running(self, host_groups: list[str] = None):
        stopped_instances = []
        logger.info("Getting instance states...")

        for group, group_hosts in self.hosts.items():
            if host_groups and group not in host_groups:
                continue

            for instance in group_hosts:
                state = instance.current_state

                if state not in (ComputeInstanceState.RUNNING, ComputeInstanceState.STOPPED):
                    raise RuntimeError(f'Host "{instance}" is in invalid state: {state}')

                if state == ComputeInstanceState.STOPPED:
                    stopped_instances.append(instance.instance_id)

        provider = self.project.provider
        if stopped_instances:
            logger.info(f"Starting {len(stopped_instances)} stopped instances...")
            provider.start_compute_instance(stopped_instances, wait=True)

        yield
        if stopped_instances:
            logger.info("Restoring instances states...")
            provider.stop_compute_instance(stopped_instances, wait=False)

    @contextmanager
    def ansible_host_inventory(self, host_groups: list[str]) -> Generator[str, None, None]:
        t = tempfile.NamedTemporaryFile(mode="w+", suffix=".yml")
        stream = t.__enter__()

        host_inventory = {}

        for group, instances in self.hosts.items():
            if host_groups and group not in host_groups:
                continue

            proxy_command = self.ssh_proxy_command(group)
            ansible_ssh_common_args = "-o StrictHostKeyChecking=no -o ConnectTimeout=60"
            if proxy_command:
                ansible_ssh_common_args = f"{ansible_ssh_common_args} -o ProxyCommand='{proxy_command}'"

            host_inventory[group] = {
                "hosts": {
                    self.get_ssh_host(instance): {
                        "ansible_ssh_common_args": ansible_ssh_common_args,
                        "ansible_python_interpreter": "/usr/bin/python3",
                    }
                    for instance in instances
                }
            }

        yaml.dump(host_inventory, stream)
        yield stream.name
        t.__exit__(None, None, None)

    @contextmanager
    def ansible_extra_vars(self) -> Generator[str, None, None]:
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".yml") as t:
            context = self.environment.context.as_dict()
            # Remove 'images' key as it is non-picklable
            context.pop("images", None)
            yaml.dump(context, t)
            yield t.name



import os
import subprocess
from typing import TYPE_CHECKING

from strong_opx.hcl.runner import HCLRunner

if TYPE_CHECKING:
    from strong_opx.project import Environment


class PackerRunner(HCLRunner):
    extension = ".pkr.hcl"
    env_var_prefix = "PKR_VAR"

    def get_executable(self) -> str:
        return self.environment.project.config.packer_executable

    def _run(self, args: tuple[str, ...], env: dict[str, str], **kwargs) -> subprocess.CompletedProcess:
        env["PKR_PLUGIN_PATH"] = os.path.join(self.environment.project.path, ".packer")
        return super()._run(args, env, **kwargs)


def run_packer(environment: "Environment", command: str, *additional_args: str):
    runner = PackerRunner(environment=environment, directory=os.path.join(environment.project.path, "packer"))
    runner.run(command, additional_args=additional_args)



import argparse
from typing import Any

from strong_opx.exceptions import CommandError
from strong_opx.management.command import ConnectCommand
from strong_opx.utils.shell import shell


def parse_path(path):
    if path[0] == "@":
        return True, path[1:]

    return False, path


class Command(ConnectCommand):
    help_text = "SCP into private host inside a project/environment"
    allow_additional_args = True
    examples = [
        "strong-opx scp --project <project> --env <env> 10.0.0.1 @<remote-path> <local-path>",
        "strong-opx scp --project <project> --env <env> 10.0.0.1 <local-path> @<remote-path>",
        "strong-opx scp --project <project> --env <env> primary @<remote-path> <local-path>",
        "strong-opx scp --project <project> --env <env> primary:2 <remote-path> @<local-path>",
    ]

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument("source", help="Source path. To represent server path prepend @")
        parser.add_argument("target", help="Target path. To represent server path prepend @")

    def shell(self, args: list[str], options: dict[str, Any]) -> None:
        source_remote, source_path = parse_path(options["source"])
        target_remote, target_path = parse_path(options["target"])

        if source_remote == target_remote:
            raise CommandError("source and target are referring to remote machine")

        remote_host = args.pop(2)

        if source_remote:
            source_path = f"{remote_host}:{source_path}"

        if target_remote:
            target_path = f"{remote_host}:{target_path}"

        args.append(source_path)
        args.append(target_path)
        shell("scp " + " ".join(args), shell=True)



import json
import os.path
import tempfile
from functools import cached_property
from subprocess import CompletedProcess
from typing import TYPE_CHECKING, Optional, Union

from pydantic import AnyUrl, Field
from pydantic.dataclasses import dataclass

from strong_opx.exceptions import HelmError
from strong_opx.template import FileTemplate
from strong_opx.utils.prompt import input_boolean
from strong_opx.utils.shell import shell

if TYPE_CHECKING:
    from strong_opx.platforms import KubernetesPlatform

HELM_DEFAULT_REPO_NAME = "stable"
HELM_DEFAULT_REPO_URL = "https://charts.helm.sh/stable"


class HelmChartQualifiedName(str):
    def __new__(cls, namespace: str, name: str):
        instance = super().__new__(cls, f"{namespace}:{name}")
        instance.name = name
        instance.namespace = namespace
        return instance


@dataclass
class HelmChart:
    repo: str
    chart: str
    name: Optional[str] = None
    version: Optional[str] = None
    namespace: str = "default"
    environment: Optional[Union[list[str], str]] = None
    values: Optional[str] = None
    timeout: str = "5m0s"

    def __post_init__(self):
        if self.name is None:
            self.name = self.chart

    @cached_property
    def qualified_name(self) -> HelmChartQualifiedName:
        return HelmChartQualifiedName(self.namespace, self.name)


@dataclass
class HelmConfig:
    charts: list[HelmChart] = Field(default_factory=list)
    repos: dict[str, AnyUrl] = None

    def __post_init__(self):
        if self.repos is None:
            self.repos = {}
        else:
            self.repos = {k: str(v) for k, v in self.repos.items()}

        if HELM_DEFAULT_REPO_NAME not in self.repos:
            self.repos[HELM_DEFAULT_REPO_NAME] = HELM_DEFAULT_REPO_URL

        for repo, repo_url in self.repos.items():
            if not any(repo_url.startswith(prefix) for prefix in ("http://", "https://", "oci://")):
                raise HelmError(f"Invalid repo URL {repo_url} for {repo}. Must start with http://, https:// or oci://")


class HelmManager:
    def __init__(self, platform: "KubernetesPlatform"):
        self.platform = platform
        self.config_path = os.path.join(self.platform.project.path, "helm.yml")
        self.config = platform.project.helm_config or HelmConfig()

    @cached_property
    def installed_helm_charts(self) -> list[HelmChartQualifiedName]:
        output = json.loads(self.run("list", "--output", "json", "-A", capture_output=True).stdout.decode("utf8"))
        return [HelmChartQualifiedName(helm["namespace"], helm["name"]) for helm in output]

    def run(self, *additional_args: str, **kwargs) -> CompletedProcess:
        self.platform.configure_kubernetes()
        return shell(("helm", "--kubeconfig", self.platform.kube_config_path) + additional_args, **kwargs)

    def apply(self, upgrade: bool, charts: list[str] = None, additional_args: tuple[str, ...] = ()) -> None:
        if upgrade:
            cmd_prefix = ("upgrade", "--install")
        else:
            cmd_prefix = ("install",)

        for chart in self.config.charts:
            if not upgrade and chart.qualified_name in self.installed_helm_charts:
                continue

            if charts and chart.name not in charts:
                continue

            starting_args = [
                chart.name,
                chart.chart,
                "--repo",
                str(self.config.repos[chart.repo]),
            ]

            if str(self.config.repos[chart.repo]).startswith("oci://"):
                starting_args = [
                    chart.name,
                    str(self.config.repos[chart.repo]),
                ]

            args = [
                *starting_args,
                "-n",
                chart.namespace,
                "--create-namespace",
                "--wait",
                "--timeout",
                str(chart.timeout),
            ]

            if chart.version:
                args.append("--version")
                args.append(chart.version)

            if chart.values:
                file_path = os.path.join(self.platform.project.path, chart.values)
                if not os.path.exists(file_path):
                    raise HelmError(f"Unable to locate specified values {chart.values} for {chart.name}")

                with tempfile.NamedTemporaryFile(suffix=f"_{os.path.basename(file_path)}") as f:
                    FileTemplate(file_path).render_to_file(f.name, self.platform.environment.context)

                    args.append("--values")
                    args.append(f.name)
                    self.run(*cmd_prefix, *args, *additional_args)
            else:
                self.run(*cmd_prefix, *args, *additional_args)

    def prune(self, additional_args: list[str] = ()) -> bool:
        to_uninstall = [
            chart_name
            for chart_name in self.installed_helm_charts
            if all(chart.qualified_name != chart_name for chart in self.config.charts)
        ]

        if not to_uninstall:
            return False

        print("Following releases will be uninstalled")
        print("  {}".format("\n  ".join(to_uninstall)))

        if input_boolean("Do you want to continue", default=False):
            for helm in to_uninstall:
                self.run("uninstall", helm.name, "-n", helm.namespace, *additional_args)



import os
from argparse import ArgumentTypeError
from typing import Generator, Optional

from strong_opx.config import PROJECT_CONFIG_FILE, system_config
from strong_opx.exceptions import CommandError
from strong_opx.project import Environment, Project
from strong_opx.utils.prompt import select_prompt


def validate_project_name(value: str) -> str:
    registered_projects = system_config.registered_projects
    if value not in registered_projects:
        raise ArgumentTypeError(f'Unknown project: {value}, Registered projects: {", ".join(registered_projects)}')

    return value


def select_project(project_name: str) -> Project:
    if not project_name:
        project = get_current_project()
        if project:
            return project

    if not project_name:
        project_name = os.getenv("STRONG_OPX_PROJECT")

    if not project_name:
        project_name = select_prompt("Select Project", sorted(system_config.registered_projects))

    return Project.from_name(project_name)


def select_environment(project: Project, environment_name: str) -> Environment:
    if environment_name is None:
        if not project.environments:
            raise CommandError("Project has no environment")
        if len(project.environments) == 1:
            environment_name = project.environments[0]
        else:
            environment_name = os.getenv("STRONG_OPX_ENVIRONMENT")

        if not environment_name:
            environment_name = select_prompt("Select Environment", sorted(project.environments))

    return project.select_environment(name=environment_name)


def walk_to_root() -> Generator[str, None, None]:
    current_dir = os.path.abspath(os.getcwd())

    last_dir = None
    while last_dir != current_dir:
        yield current_dir
        parent_dir = os.path.abspath(os.path.join(current_dir, os.path.pardir))
        last_dir, current_dir = current_dir, parent_dir


def find_project_config_path() -> Optional[str]:
    for directory_path in walk_to_root():
        project_config = os.path.join(directory_path, PROJECT_CONFIG_FILE)
        if os.path.exists(project_config) and os.path.isfile(project_config):
            return project_config


def get_current_project() -> Optional[Project]:
    project_config_path = find_project_config_path()
    if not project_config_path:
        return

    project = Project.from_config(project_config_path)
    if project.name not in system_config.registered_projects:
        raise CommandError(f"Current directory holds project {project.name} that is unknown to registry")

    expected_project_path = system_config.get_project_path(project.name)
    if expected_project_path != os.path.dirname(project_config_path):
        raise CommandError(
            f"Current directory holds project {project.name} "
            f"but another project in different location is registered by that name"
        )

    return project



import argparse
import os
from typing import Any

from strong_opx.management.command import ProjectCommand
from strong_opx.platforms import KubernetesPlatform
from strong_opx.project import Environment


class Command(ProjectCommand):
    help_text = "Manage Kubernetes Cluster"
    allow_additional_args = True
    parse_known_args = True

    examples = [
        "strong-opx kubectl --project <project> --env <env> -- apply -f <file>.yml",
    ]

    def add_arguments(self, parser: argparse.ArgumentParser):
        super().add_arguments(parser)
        parser.add_argument("--update-config", action="store_true", default=False, help="Force update kubectl config")

    def handle(self, environment: Environment, additional_args: tuple[str, ...], update_config: bool, **options: Any):
        platform = environment.select_platform(KubernetesPlatform)

        if update_config and os.path.exists(platform.kube_config_path):
            os.remove(platform.kube_config_path)

        platform.kubectl(*additional_args)



from typing import TYPE_CHECKING, Union

from pydantic_core import core_schema

from strong_opx.template import ObjectTemplate

if TYPE_CHECKING:
    from strong_opx.project import Environment

T_VARIABLE_CONFIG = Union[
    str,
    list[Union[str, dict[str, Union[str, list[str]]]]],
    dict[str, Union[str, list[str]]],
]


class VariableConfig:
    value: T_VARIABLE_CONFIG

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    def get_paths(self, environment: "Environment") -> list[str]:
        if isinstance(self.value, str):
            paths = [self.value]
        elif isinstance(self.value, dict):
            paths = self.value.get(environment.name) or []
            if isinstance(paths, str):
                paths = [paths]
        else:
            paths = []
            for path_config in self.value:
                if isinstance(path_config, dict):
                    path_config = path_config.get(environment.name) or []

                if isinstance(path_config, str):
                    paths.append(path_config)
                else:
                    paths.extend(path_config)

        return ObjectTemplate(environment.base_context).render(paths)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):
        return core_schema.chain_schema(
            [
                handler(T_VARIABLE_CONFIG),
                core_schema.no_info_plain_validator_function(source_type),
            ]
        )



import argparse
import logging
import os
from typing import Any, Optional

from strong_opx.exceptions import CommandError
from strong_opx.management.command import ProjectCommand
from strong_opx.project import Environment, Project
from strong_opx.providers.container_registry import (
    DEFAULT_DOCKER_TAG,
    AbstractContainerRegistry,
    current_container_registry,
)
from strong_opx.utils.shell import shell, ssh_agent

logger = logging.getLogger(__name__)


def docker_tag_string(value: Optional[str]) -> str:
    value = value and value.strip()
    if not value:
        return DEFAULT_DOCKER_TAG

    return value


def get_ecr_tags_to_apply(
    registry: "AbstractContainerRegistry",
    environment: "Environment",
    repository_name: str,
    docker_tag: list[str],
) -> tuple[str, set[str]]:
    repository_uri = registry.get_or_create_repository_uri(repository_name)

    revision = registry.get_latest_revision(repository_name)
    image_tag = registry.tag_from_revision(revision + 1)

    tags = {
        f"{repository_uri}:{image_tag}",
        f"{repository_uri}:{DEFAULT_DOCKER_TAG}",
        f"{repository_uri}:{environment.name}-{DEFAULT_DOCKER_TAG}",
    }
    for tag in docker_tag:
        tags.add(f"{repository_uri}:{tag}")

    return repository_uri, tags


def docker_build(
    project: Project, tags: set[str], path: str, additional_args: tuple[str, ...], mount_ssh: bool = False
) -> None:
    cmd = [project.config.docker_executable, "buildx", "build"]
    for tag in sorted(tags):
        cmd.append("-t")
        cmd.append(tag)

    if mount_ssh:
        cmd.append("--ssh")
        cmd.append("default")

    cmd.extend(additional_args)
    cmd.append(path)
    shell(cmd)


class Command(ProjectCommand):
    allow_additional_args = True

    def add_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument("path", help="Docker build context")
        parser.add_argument("--name", help="Image name. Default: directory name")
        parser.add_argument(
            "--ssh",
            dest="mount_ssh",
            action="store_true",
            default=False,
            help="Run ssh-agent and add ssh key to the build context",
        )
        parser.add_argument(
            "--ssh-key",
            dest="ssh_key",
            help="When --ssh is used, use this key instead of the default",
        )
        parser.add_argument(
            "--push",
            dest="docker_push",
            action="store_true",
            default=False,
            help="Push built image to ECR",
        )
        parser.add_argument(
            "--build-arg",
            dest="build_args",
            nargs="+",
            help="Additional build args to docker-build (values will be auto-resolved from stored vars)",
        )
        parser.add_argument(
            "--tag",
            dest="docker_tags",
            type=docker_tag_string,
            nargs="+",
            default=[],
            help=f"Additional tags to apply. {DEFAULT_DOCKER_TAG} tag will always be included",
        )
        super().add_arguments(parser)

    def handle(
        self,
        path: str,
        project: Project,
        environment: Environment,
        additional_args: tuple[str, ...],
        docker_tags: list[str],
        mount_ssh: bool = False,
        ssh_key: Optional[str] = None,
        docker_push: bool = False,
        name: Optional[str] = None,
        build_args: list[str] = None,
        **options: Any,
    ):
        registry = current_container_registry(environment)
        if registry is None:
            raise CommandError("No container registry configured for current provider")

        if not os.path.isabs(path):
            if os.path.exists(os.path.abspath(path)):
                path = os.path.abspath(path)
            else:
                path = os.path.join(environment.project.path, path)

        if not os.path.exists(path):
            raise CommandError(f"Specified path does not exists: {path}")

        if name is None:
            name = os.path.basename(path)

        if build_args:
            args = []
            for k, v in environment.context.require(*build_args).items():
                args.append("--build-arg")
                args.append(f"{k}={v}")

            additional_args += tuple(args)

        if docker_push:
            registry.login()
            repository_uri, tags = get_ecr_tags_to_apply(registry, environment, name, docker_tags)
            additional_args += (
                "--push",
                "--cache-to",
                "type=inline",
                "--cache-from",
                f"{repository_uri}:{DEFAULT_DOCKER_TAG}",
            )
        else:
            tags = {f"{name}:{tag}" for tag in docker_tags}
            tags.add(f"{name}:{DEFAULT_DOCKER_TAG}")

        if mount_ssh:
            with ssh_agent(ssh_key or project.config.git_ssh_key):
                docker_build(project, tags, path, additional_args, mount_ssh=True)
        else:
            docker_build(project, tags, path, additional_args)



import argparse
import os
from typing import Any

from strong_opx.management.command import ProjectCommand
from strong_opx.platforms import KubernetesPlatform
from strong_opx.project import Environment


class Command(ProjectCommand):
    help_text = "Access Kubernetes Cluster Resources"
    parse_known_args = True

    examples = [
        "strong-opx k8s --project <project> --env <env> dashboard up",
    ]

    def add_arguments(self, parser: argparse.ArgumentParser):
        super().add_arguments(parser)
        parser.add_argument("--update-config", action="store_true", default=False, help="Force update kubectl config")

        parser.add_argument("plugin", help="Plugin to execute")
        parser.add_argument("operation", help="Plugin operation to execute")

    def handle(
        self,
        environment: Environment,
        update_config: bool,
        plugin: str,
        operation: str,
        additional_args: tuple[str, ...],
        **options: Any,
    ):
        platform = environment.select_platform(KubernetesPlatform)

        if update_config and os.path.exists(platform.kube_config_path):
            os.remove(platform.kube_config_path)

        if plugin is None:
            print("Available plugins:")
            for plugin_name in platform.plugins:
                print(f"  {plugin_name}")
            return

        platform.plugin(plugin).run(operation, *additional_args)



import warnings
from typing import Annotated

from pydantic import BaseModel, Field, ValidationError, create_model

from strong_opx import yaml
from strong_opx.config import StrongOpxConfig
from strong_opx.helm import HelmConfig
from strong_opx.project.vars import VariableConfig
from strong_opx.providers import Provider, SecretProvider, select_provider
from strong_opx.utils.validation import translate_pydantic_errors


class ProjectConfig(BaseModel):
    name: str
    dirname: str = None

    provider: Provider
    strong_opx: StrongOpxConfig = None
    helm: HelmConfig = None

    secret: SecretProvider
    vars: VariableConfig

    @classmethod
    def from_file(cls, config_path: str) -> "ProjectConfig":
        config_dict = yaml.load(config_path)
        if config_dict.pop("version", None) is not None:
            warnings.warn(f"`version` is deprecated. Remove this from {config_path} to dismiss this warning")

        provider_name, provider_class = select_provider(config_dict)

        project_config_cls = create_model(
            "GeneratedProjectConfig",
            provider=Annotated[provider_class, Field(alias=provider_name)],
            __base__=cls,
        )

        try:
            return project_config_cls(**config_dict)
        except ValidationError as ex:
            raise translate_pydantic_errors(config_dict, ex)



import importlib
import logging
from importlib import import_module

logger = logging.getLogger(__name__)


def import_string(dotted_path: str):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    try:
        module = import_module(module_path, class_name)
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute/class') from err


def import_module_attr_if_exists(module_name: str, attr_name: str):
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        if e.name != module_name:
            raise

        return None

    try:
        return getattr(module, attr_name)
    except AttributeError:
        logger.warning(f"{attr_name} not found in {module_name}. Please check the module.")



import os
import signal
from functools import cached_property
from subprocess import CompletedProcess, Popen
from typing import TYPE_CHECKING, Any, Optional

import filelock
from pydantic import BaseModel

from strong_opx import yaml
from strong_opx.config import system_config
from strong_opx.exceptions import CommandError
from strong_opx.platforms.base import Platform
from strong_opx.platforms.deployments import KubeCtlDeploymentProvider
from strong_opx.platforms.plugins import KubernetesDashboardPlugin
from strong_opx.providers import current_container_registry
from strong_opx.utils.shell import shell
from strong_opx.utils.socket import get_free_tcp_port

if TYPE_CHECKING:
    from strong_opx.template import Context


class KubernetesPlatformConfig(BaseModel):
    cluster_name: str
    service_role: str = None


class KubernetesPlatform(Platform):
    config_class = KubernetesPlatformConfig
    config_namespace = "kubernetes"

    if TYPE_CHECKING:
        cluster_name: str
        service_role: Optional[str]

    deployment_providers = (KubeCtlDeploymentProvider,)
    plugins = {
        "dashboard": KubernetesDashboardPlugin,
    }

    def init_context(self, context: "Context") -> None:
        context["images"] = current_container_registry(self.environment)

    @cached_property
    def kube_config_path(self) -> str:
        return os.path.join(
            system_config.get_project_config_dir(self.project.name),
            f"kubeconfig-{self.environment.name}",
        )

    def configure_kubernetes(self) -> None:
        provider = self.project.provider

        if self.service_role is not None:
            provider.assume_service_role(self.service_role)

        if not os.path.exists(self.kube_config_path):
            provider.update_kubeconfig(self.cluster_name, self.kube_config_path)
            self._post_process_kubeconfig()

    def kubectl(self, *additional_args: str, **kwargs: Any) -> CompletedProcess:
        self.configure_kubernetes()
        return shell(
            (self.project.config.kubectl_executable, "--kubeconfig", self.kube_config_path) + additional_args,
            **kwargs,
        )

    @cached_property
    def proxy_status_file_path(self) -> str:
        return os.path.join(
            system_config.get_project_config_dir(self.project.name),
            f"kubernetes-proxy-{self.environment.name}",
        )

    def get_proxy_status(self) -> Optional[tuple[int, int]]:
        if not os.path.exists(self.proxy_status_file_path):
            return None

        with open(self.proxy_status_file_path, "r") as f:
            pid = f.readline().strip()
            port = f.readline().strip()

            pid = int(pid) if pid else None
            port = int(port)

        try:
            # `os.kill` will not actually kill the process.
            # (Source: https://stackoverflow.com/a/7647264)
            os.kill(pid, 0)
        except ProcessLookupError:
            os.remove(self.proxy_status_file_path)
        else:
            return pid, port

    def save_proxy_status(self, pid: int, port: int) -> None:
        with open(self.proxy_status_file_path, "w") as f:
            f.write(str(pid))
            f.write("\n")
            f.write(str(port))

    def proxy_lock(self) -> filelock.BaseFileLock:
        return filelock.FileLock(self.proxy_status_file_path + ".lock")

    def start_proxy(self, detached: bool) -> None:
        process = None
        with self.proxy_lock():
            status = self.get_proxy_status()
            if status is not None:
                _, port = status
                print(f"Kubernetes proxy is already running on port {port}")
            else:
                port = get_free_tcp_port()
                print(f"Starting kubernetes proxy on port {port}")

                self.configure_kubernetes()
                process = Popen(
                    (
                        self.project.config.kubectl_executable,
                        "--kubeconfig",
                        self.kube_config_path,
                        "proxy",
                        "-p",
                        str(port),
                    ),
                    start_new_session=True,
                )
                self.save_proxy_status(process.pid, port)

        if process and not detached:
            process.wait()

    def stop_proxy(self) -> bool:
        status = self.get_proxy_status()
        if status:
            pid, _ = status
            if pid is not None:
                os.killpg(os.getpgid(pid), signal.SIGTERM)
                return True

        return False

    def ensure_proxy_is_running(self) -> None:
        with self.proxy_lock():
            status = self.get_proxy_status()
            if not status:
                raise CommandError("Kubernetes proxy is not running")

    def _post_process_kubeconfig(self):
        # Remove environment variables, those will be set again automatically by strong-opx when executing
        # kubectl commands.

        kube_config = yaml.load(self.kube_config_path)
        for user in kube_config["users"]:
            user_exec = user["user"]["exec"]
            user_exec.pop("env", None)

        yaml.dump(kube_config, self.kube_config_path)



try:
    from enum import StrEnum
except ImportError:
    from enum import Enum

    class StrEnum(str, Enum):
        def __new__(cls, value):
            if not isinstance(value, str):
                raise TypeError(f"Value {value!r} is not a string")

            obj = str.__new__(cls, value)
            obj._value_ = value
            return obj

        def __str__(self):
            return self._value_



"""
Cython generator doesn't work with PyYAML's custom constructors.
This module isn't Cythonized and only holds the custom constructors for loading list and dict
"""

import yaml

from strong_opx.utils.tracking import OpxMapping, OpxSequence, set_position_from_yaml_mark


def construct_yaml_seq(loader: yaml.SafeLoader, node: yaml.SequenceNode):
    value = OpxSequence()
    set_position_from_yaml_mark(value, node.start_mark, node.end_mark)
    yield value
    value.extend(loader.construct_sequence(node))


def construct_yaml_map(loader: yaml.SafeLoader, node: yaml.MappingNode):
    value = OpxMapping()
    set_position_from_yaml_mark(value, node.start_mark, node.end_mark)
    yield value
    value.update(loader.construct_mapping(node))



import argparse
import importlib.util
import os
import pkgutil
import sys
from difflib import SequenceMatcher
from typing import Generator

import colorama

from strong_opx import __version__
from strong_opx.management.utils import validate_project_name

colorama.init()

COMMAND_ALIAS = {
    "g": "generate",
}


def _command_name_from_module_name(module_name: str) -> str:
    return module_name.rsplit(".", 1)[-1].replace("_", ":")


def _list_commands() -> Generator[str, None, None]:
    from strong_opx.management import commands

    for _, module_name, _ in pkgutil.iter_modules(commands.__path__):
        yield _command_name_from_module_name(module_name)


def _list_similar_commands(user_command: str) -> list[str]:
    similarity = {}

    for known_command in _list_commands():
        ratio = SequenceMatcher(None, known_command, user_command).ratio()
        if ratio > 0.8:
            similarity[known_command] = ratio

    return [x[0] for x in sorted(similarity.items(), key=lambda x: x[1])]


def _validate_command(command: str):
    command = COMMAND_ALIAS.get(command, command)
    module_name = command.replace(":", "_").replace("-", "_")

    try:
        module_spec = importlib.util.find_spec(f"strong_opx.management.commands.{module_name}")
    except ModuleNotFoundError:
        module_spec = None

    if module_spec is None:
        print(f'strong-opx: "{command}" is not a strong-opx command. See strong-opx --help')

        similar_commands = _list_similar_commands(command)
        if similar_commands:
            print("\nThe most similar command(s) are:")
            for command_name in similar_commands:
                print(f"    {command_name}")

            print()

        exit(1)

    return module_spec


def _print_main_help(parser: argparse.ArgumentParser):
    print(f'\nType "{parser.prog} <command> --help" for help on specific command\n')
    print("strong-opx commands:")

    for command_name in _list_commands():
        print(f"    {command_name}")


def main():
    parser = argparse.ArgumentParser(usage="%(prog)s [--help] subcommand [options] [args]", add_help=False)
    parser.add_argument("-h", "--help", action="store_true", help="show this help message and exit")
    parser.add_argument("--version", action="store_true", help="show strong-opx version and exit")
    parser.add_argument("--project", type=validate_project_name, help="Select project by its name.")
    parser.add_argument("--env")
    parser.add_argument(dest="command", nargs="?", help="name of command to execute", type=_validate_command)

    argv = sys.argv[1:]
    try:
        i = argv.index("--")
        additional_args = tuple(argv[i + 1 :])
        argv = argv[:i]
    except ValueError:
        additional_args = None

    args, command_args = parser.parse_known_args(args=argv)
    if additional_args:
        command_args.append("--")
        command_args.extend(additional_args)

    if args.command:
        if args.help:
            command_args.append("--help")
    else:
        if args.help:
            _print_main_help(parser)
        elif args.version:
            print(__version__)
        else:
            parser.error("A subcommand is required, e.g., strong-opx deploy")
            exit(1)

        exit(0)

    if args.env:
        command_args.insert(0, "--env")
        command_args.insert(1, args.env)

    if args.project:
        command_args.insert(0, "--project")
        command_args.insert(1, args.project)

    path_env = os.environ.get("PATH", "")
    os.environ["PATH"] = f"{os.path.dirname(sys.executable)}:{path_env}"

    command_name = _command_name_from_module_name(args.command.name)
    command_module = args.command.loader.load_module()
    command_class = getattr(command_module, "Command", None)

    if command_class is None:
        raise ValueError(f"{args.command.origin} does not define a Command class")

    command = command_class()
    command_args.insert(0, parser.prog)
    command_args.insert(1, command_name)
    command.run_from_argv(command_args)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING, Callable

from strong_opx.providers.azure.config import AzureConfig
from strong_opx.providers.azure.context_hooks import update_environ_hook
from strong_opx.providers.provider import Provider

if TYPE_CHECKING:
    from strong_opx.template import Context


class AzureProvider(Provider):
    config: AzureConfig

    def get_additional_context_hooks(self) -> tuple[Callable[["Context"], None], ...]:
        return (
            self.update_context,
            update_environ_hook,
        )

    def update_context(self, context: "Context"):
        context.update(self.config.dict())



import ast
from dataclasses import dataclass
from typing import Any, Generator, Optional

from strong_opx.template.registry import SAFE_BUILTINS
from strong_opx.utils.tracking import Position

REF_SEP = "."


@dataclass(frozen=True)
class VariableRef:
    """
    Hold information about a variable reference in the code.

    :param name: The full name of the variable reference nodes separated by REF_SEP
    :param nodes: The individual nodes of the variable reference
    """

    name: str
    nodes: tuple[str, ...]

    start_pos: Position
    end_pos: Position

    @classmethod
    def from_ast_node(cls, top_node: ast.AST) -> Optional["VariableRef"]:
        """
        Extract a variable reference from an AST node. If the node is not a variable reference, None is returned.

        `VariableRef.nodes` are extracted using following rules:
        - If ast node is of type `ast.Name`, node is the name of the variable
        - If ast node is of type `ast.Attribute`, it's the nodes from value of the `ast.Attribute` joined with the
          attribute name of the `ast.Attribute`
            - nodes from value of the `ast.Attribute` are extracted recursively
            - If the value of the `ast.Attribute` has no nodes, ast node is skipped
        - If ast node is of type `ast.Subscript`, it's the nodes from value of the `ast.Subscript` joined with the
          slice value of the `ast.Subscript`
            - nodes from value of the `ast.Subscript` are extracted recursively
            - If the value of the `ast.Subscript` has no nodes, ast node is skipped
            - If slice value of the `ast.Subscript` is not a `ast.Constant`, ast `node.slice` is skipped
        - If ast node is of any other type, ast node is skipped.

        `VariableRef.name` is the nodes joined with REF_SEP
        `VariableRef.start_pos` is the start position of the top node
        `VariableRef.end_pos` is the end position of the top node. If the top node is a `ast.Subscript` and the
        slice value is not a `ast.Constant`, end position is the end position of the value node.

        :param top_node: The top node to extract the variable reference from
        :return VariableRef: The variable reference if found, None otherwise
        """
        start_position = Position(top_node.lineno, top_node.col_offset + 1)
        end_position = Position(top_node.end_lineno, top_node.end_col_offset + 1)

        def extract_ref_nodes(node: ast.AST) -> Generator[str, None, None]:
            if isinstance(node, ast.Name):
                yield node.id

            elif isinstance(node, ast.Attribute):
                value = tuple(extract_ref_nodes(node.value))  # NOQA
                if value:
                    yield from value
                    yield node.attr

            elif isinstance(node, ast.Subscript):
                value = tuple(extract_ref_nodes(node.value))  # NOQA
                if value:
                    yield from value
                    if isinstance(node.slice, ast.Constant):
                        yield str(node.slice.value)  # NOQA
                    else:
                        nonlocal end_position
                        end_position = Position(node.value.end_lineno, node.value.end_col_offset + 1)

        nodes = tuple(extract_ref_nodes(top_node))
        if not nodes:
            return None

        return VariableRef(name=REF_SEP.join(nodes), nodes=nodes, start_pos=start_position, end_pos=end_position)


class VariableStore:
    """
    A store for variables in the Template code. It keeps track of the variables declared in the code and their
    references.

    Store keeps track of the following:
    - Local variables declared in the current scope
    - Global variables declared in the code
    - References to context variables in the code
    """

    def __init__(self):
        self.locals: list[str] = []
        self.locals_bound: list[int] = [len(self.locals)]
        self.globals = {
            "__builtins__": SAFE_BUILTINS,
        }

        self.refs: dict[str, VariableRef] = {}

    def get_ref(self, name: str) -> Optional[VariableRef]:
        """
        Get the best matching variable reference that starts with `name`. If there are multiple matching
        references one with the lowest nodes will be returned.

        :param name: `REF_SEP` seperated name of the variable reference
        :return: The variable reference if found, None otherwise
        """

        nodes = tuple(name.split(REF_SEP))
        n_nodes = len(nodes)

        matching_refs = [ref for ref in self.refs.values() if ref.nodes[:n_nodes] == nodes]
        if matching_refs:
            return sorted(matching_refs, key=lambda ref: len(ref.nodes))[0]

    def begin_scope(self) -> None:
        self.locals_bound.append(len(self.locals))

    def end_scope(self) -> None:
        bound = self.locals_bound.pop()
        self.locals = self.locals[:bound]

    def declare(self, name: str) -> None:
        """
        Declare a local variable in the current scope
        :param name: The name of the variable
        """
        self.locals.append(name)

    def define(self, name: str, value: Any) -> None:
        """
        Define a global variable in the store. This is used to define additional variables in the store that are not
        present in the code. For example, template filters.

        :param name: The name of the variable
        :param value: The value of the variable
        """
        self.globals[name] = value

    def __contains__(self, name) -> bool:
        """
        Check if a variable is defined in the store.
        This checks if the variable is defined in the locals or the globals or allowed builtins.

        :param name: The name of the variable
        :return: True if the variable is defined, False otherwise
        """

        return name in self.locals or name in self.globals or name in SAFE_BUILTINS



from typing import Annotated

from packaging.specifiers import SpecifierSet
from pydantic.dataclasses import dataclass
from pydantic.functional_validators import PlainValidator
from pydantic_core import CoreSchema, core_schema

from strong_opx import __version__ as strong_opx_version
from strong_opx.exceptions import ImproperlyConfiguredError

DEFAULT_TEMPLATING_ENGINE = "strong_opx"


def validate_template_engine(value: str) -> str:
    if value not in ("strong_opx", "jinja2"):
        raise ValueError(f'Unknown templating engine: "{value}". Allowed values: strong_opx, jinja2')

    return value


@dataclass
class StrongOpxConfig:
    templating_engine: Annotated[str, PlainValidator(validate_template_engine)] = DEFAULT_TEMPLATING_ENGINE
    required_version: Annotated[SpecifierSet, PlainValidator(SpecifierSet)] = None

    def update(self, config: "StrongOpxConfig") -> None:
        self.templating_engine = config.templating_engine
        self.required_version = config.required_version

        if self.required_version and not self.required_version.contains(strong_opx_version):
            raise ImproperlyConfiguredError(
                f"Project requires strong-opx{self.required_version} "
                f"but strong-opx=={strong_opx_version} is installed"
            )

    # TODO: Remove that as we remove Cython
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler) -> CoreSchema:
        templating_engine_schema = core_schema.with_default_schema(
            core_schema.chain_schema(
                [
                    core_schema.str_schema(),
                    core_schema.no_info_plain_validator_function(validate_template_engine),
                ]
            ),
            default=DEFAULT_TEMPLATING_ENGINE,
        )

        required_version_schema = core_schema.with_default_schema(
            core_schema.chain_schema(
                [
                    core_schema.str_schema(),
                    core_schema.no_info_plain_validator_function(SpecifierSet),
                ]
            ),
            default=None,
        )

        return core_schema.dataclass_schema(
            cls,
            core_schema.dataclass_args_schema(
                cls.__name__,
                [
                    core_schema.dataclass_field(
                        "templating_engine",
                        templating_engine_schema,
                        kw_only=True,
                    ),
                    core_schema.dataclass_field(
                        "required_version",
                        required_version_schema,
                        kw_only=True,
                    ),
                ],
            ),
            ["templating_engine", "required_version"],
        )



import os
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from dateutil.parser import parse as parse_dt

from strong_opx.config import CONFIG_DIR, Config

AWS_CREDENTIALS_FILE = os.path.join(os.path.expanduser("~"), ".aws", "credentials")


def mfa_profile_name(profile_name: str) -> str:
    return f"{profile_name}--mfa"


@dataclass
class AWSCredential:
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str

    def dict(self):
        return {
            "AWS_ACCESS_KEY_ID": self.aws_access_key_id,
            "AWS_SECRET_ACCESS_KEY": self.aws_secret_access_key,
            "AWS_SESSION_TOKEN": self.aws_session_token,
        }


class AWSCredentialConfig(Config):
    def __init__(self, config_path: str = None):
        super().__init__(config_path or AWS_CREDENTIALS_FILE)

    def are_credential_expired(self, profile_name: str) -> bool:
        expiry = self.get(profile_name, "expires", fallback=None)
        if not expiry:
            return True

        expiry_dt = parse_dt(expiry)
        if expiry_dt and datetime.now() < expiry_dt:
            return False

        return True

    def get_credential(self, profile_name: str) -> Optional[AWSCredential]:
        if self.are_credential_expired(profile_name):
            return None

        return AWSCredential(
            aws_access_key_id=self.get(profile_name, "aws_access_key_id", fallback=""),
            aws_secret_access_key=self.get(profile_name, "aws_secret_access_key", fallback=""),
            aws_session_token=self.get(profile_name, "aws_session_token", fallback=""),
        )

    def set_credential(self, profile_name: str, credential: AWSCredential, expires: datetime):
        self.remove_section(profile_name)
        self.add_section(profile_name)
        self.set(profile_name, "aws_access_key_id", credential.aws_access_key_id)
        self.set(profile_name, "aws_secret_access_key", credential.aws_secret_access_key)
        self.set(profile_name, "aws_session_token", credential.aws_session_token)
        self.set(profile_name, "expires", expires.astimezone().replace(tzinfo=None).isoformat())
        self.save()


aws_credentials = AWSCredentialConfig(os.path.join(CONFIG_DIR, "aws-credentials"))



import os
from contextlib import contextmanager
from typing import Any, TextIO

import yaml

from strong_opx.exceptions import YAMLError
from strong_opx.utils.tracking import OpxFloat, OpxInteger, OpxString, Position, set_position_from_yaml_mark
from strong_opx.vault import VaultCipher
from strong_opx.yaml.cython_compat import construct_yaml_map, construct_yaml_seq


@contextmanager
def _yaml_loader(file_path: str):
    with open(file_path, "r") as f:
        loader = OpxYAMLLoader(stream=f, file_path=file_path)
        try:
            yield loader
        except yaml.YAMLError as ex:
            kwargs = {}

            if isinstance(ex, yaml.MarkedYAMLError):
                lines = [ex.problem]
                kwargs["start_pos"] = Position.from_yaml_mark(ex.problem_mark)
            else:
                lines = str(ex).splitlines()

            raise YAMLError(
                "\n".join(lines),
                file_path=file_path,
                hint="error maybe elsewhere in the file depending on the exact syntax problem",
                **kwargs,
            )
        finally:
            loader.dispose()


def load(file_path: str) -> Any:
    with _yaml_loader(file_path) as loader:
        return loader.get_single_data()


def load_all(file_path: str) -> list[Any]:
    documents = []
    with _yaml_loader(file_path) as loader:
        while loader.check_data():
            documents.append(loader.get_data())

    return documents


class OpxYAMLLoader(yaml.SafeLoader):
    def __init__(self, stream: TextIO, file_path: str):
        self.file_path = file_path
        super().__init__(stream)

    def include(self, node: yaml.ScalarNode):
        rel_path = self.construct_scalar(node)
        abs_path = os.path.join(os.path.dirname(self.file_path), rel_path)
        if not os.path.exists(abs_path):
            raise YAMLError(
                f"Included file ({rel_path}) not found at {abs_path}",
                file_path=self.file_path,
                start_pos=Position.from_yaml_mark(node.start_mark),
                end_pos=Position.from_yaml_mark(node.end_mark),
            )

        return load(abs_path)

    def vault(self, node):
        value = VaultCipher.parse(self.construct_scalar(node))
        set_position_from_yaml_mark(value, node.start_mark, node.end_mark)
        return value

    def construct_yaml_int(self, node: yaml.ScalarNode):
        value = OpxInteger(super().construct_yaml_int(node))
        set_position_from_yaml_mark(value, node.start_mark, node.end_mark)
        return value

    def construct_yaml_float(self, node: yaml.ScalarNode):
        value = OpxFloat(super().construct_yaml_float(node))
        set_position_from_yaml_mark(value, node.start_mark, node.end_mark)
        return value

    def construct_yaml_timestamp(self, node: yaml.ScalarNode):
        value = super().construct_yaml_timestamp(node)
        set_position_from_yaml_mark(value, node.start_mark, node.end_mark)
        return value

    def construct_yaml_str(self, node: yaml.ScalarNode):
        value = OpxString(self.construct_scalar(node))
        set_position_from_yaml_mark(value, node.start_mark, node.end_mark)
        return value


# YAML Tags
OpxYAMLLoader.add_constructor("tag:yaml.org,2002:int", OpxYAMLLoader.construct_yaml_int)
OpxYAMLLoader.add_constructor("tag:yaml.org,2002:float", OpxYAMLLoader.construct_yaml_float)
OpxYAMLLoader.add_constructor("tag:yaml.org,2002:timestamp", OpxYAMLLoader.construct_yaml_timestamp)
OpxYAMLLoader.add_constructor("tag:yaml.org,2002:str", OpxYAMLLoader.construct_yaml_str)
OpxYAMLLoader.add_constructor("tag:yaml.org,2002:seq", construct_yaml_seq)
OpxYAMLLoader.add_constructor("tag:yaml.org,2002:map", construct_yaml_map)

# Additional Tags
OpxYAMLLoader.add_constructor("!include", OpxYAMLLoader.include)
OpxYAMLLoader.add_constructor("!vault", OpxYAMLLoader.vault)



from ansible.parsing.vault import CIPHER_MAPPING, AnsibleVaultError, VaultSecret

from strong_opx.exceptions import VaultError

DEFAULT_CIPHER_NAME = "AES256"


class VaultCipher:
    def __init__(self, ciphertext: str, cipher_name: str = DEFAULT_CIPHER_NAME, version: str = "1.0"):
        self.version = version
        self.ciphertext = ciphertext
        self.cipher_name = cipher_name

    def __str__(self):
        header_parts = ["$STRONG_OPX_VAULT", self.version, self.cipher_name]
        header = ";".join(header_parts)

        vault_text = [header]
        vault_text += [self.ciphertext[i : i + 80] for i in range(0, len(self.ciphertext), 80)]

        return "\n".join(vault_text)

    def __call__(self) -> str:
        from strong_opx.project import Project

        return self.decrypt(Project.current().selected_environment.vault_secret)

    def decrypt(self, secret: str) -> str:
        cipher_cls = CIPHER_MAPPING[self.cipher_name]
        ciphertext = self.ciphertext.encode("utf8", errors="surrogate_or_strict")

        vault_secret = VaultSecret(secret.encode("utf8", errors="surrogate_or_strict"))

        try:
            return cipher_cls().decrypt(ciphertext, vault_secret).decode("utf8")
        except AnsibleVaultError:
            raise VaultError("Decryption failed. Did you copied from other environment?")

    @classmethod
    def encrypt(cls, value: str, secret: str, cipher_name: str = DEFAULT_CIPHER_NAME) -> "VaultCipher":
        cipher_cls = CIPHER_MAPPING[cipher_name]
        cipher_text = (
            cipher_cls()
            .encrypt(
                value.encode("utf8", errors="surrogate_or_strict"),
                VaultSecret(secret.encode("utf8", errors="surrogate_or_strict")),
            )
            .decode("utf8")
        )

        return VaultCipher(ciphertext=cipher_text, cipher_name=cipher_name)

    @classmethod
    def parse(cls, envelope: str) -> "VaultCipher":
        lines = envelope.splitlines()
        header = lines[0].strip().split(";")

        version = header[1].strip()
        cipher_name = header[2].strip()

        ciphertext = "".join(lines[1:])
        return cls(ciphertext=ciphertext, cipher_name=cipher_name, version=version)



import argparse
import warnings
from typing import TYPE_CHECKING, Any

from strong_opx.exceptions import PluginError

if TYPE_CHECKING:
    from strong_opx.platforms.base import Platform


class PlatformPlugin:
    def __init__(self, platform: "Platform"):
        self.platform = platform

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        pass

    def parse_arguments(self, *args: str) -> dict[str, Any]:
        parser = argparse.ArgumentParser()
        self.add_arguments(parser)

        options = parser.parse_args(args)
        return vars(options)

    def is_installed(self) -> bool:
        raise NotImplementedError()

    def install(self) -> None:
        raise NotImplementedError()

    def run(self, operation: str, *args: str) -> None:
        operation = operation.lower()
        if operation == "install":
            if args:
                warnings.warn("Additional passed arguments will be ignored: {}".format(" ".join(args)))

            if self.is_installed():
                raise PluginError("Plugin is already installed")

            self.install()
            return

        if not self.is_installed():
            raise PluginError("Plugin is not installed and is not available. Please install it first")

        kwargs = self.parse_arguments(operation, *args)
        self.handle(**kwargs)

    def handle(self, **kwargs):
        raise NotImplementedError()



import argparse
import itertools
import logging
import os
import shlex
import time
from subprocess import CompletedProcess
from typing import Any, Optional

from strong_opx.exceptions import CommandError
from strong_opx.management.command import ProjectCommand
from strong_opx.platforms import GenericPlatform
from strong_opx.project import Environment, Project
from strong_opx.utils.shell import shell

logger = logging.getLogger(__name__)


def remote_exec(project: Project, ssh_host: str, script: str, use_screen: bool, session: str) -> CompletedProcess:
    if use_screen:
        executable = ("screen", "-S", session)
    else:
        executable = ("bash", "-c")

    return shell(
        (
            "ssh",
            "-ti",
            project.config.ssh_key,
            f"{project.config.ssh_user}@{ssh_host}",
            *executable,
            script,
        )
    )


class Command(ProjectCommand):
    help_text = "Execute local script on remote machine"
    allow_additional_args = True

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)

        parser.add_argument("script", nargs="?", help="Script to execute")
        parser.add_argument(
            "--session", default="default", help="An arbitrary name for this session. Session name should be unique"
        )
        parser.add_argument("--attach", default=False, action="store_true", help="Restore existing session")
        parser.add_argument(
            "--no-screen",
            default=True,
            dest="use_screen",
            action="store_false",
            help="Do not run inside screen session",
        )
        parser.add_argument(
            "--sync-only", default=False, dest="sync_only", action="store_true", help="Just sync files with remote"
        )
        parser.add_argument(
            "--host",
            help="Hostname or IP on which to execute. If not specified, defaults to first hosts in environment",
        )
        parser.add_argument(
            "--context", help="Path to directory use as context. If not specified, defaults to script parent directory"
        )

    def handle(
        self,
        project: Project,
        environment: Environment,
        script: str,
        attach: bool,
        sync_only: bool,
        use_screen: bool,
        context: Optional[str],
        host: Optional[str],
        session: Optional[str],
        additional_args: tuple[str, ...],
        **options: Any,
    ):
        platform = environment.select_platform(GenericPlatform)

        if host:
            instance = platform.resolve_instance(host)[0]
        else:
            instance = next(itertools.chain.from_iterable(platform.hosts.values()), None)
            if instance is None:
                raise CommandError("No host in environment")

        ssh_host = platform.get_ssh_host(instance)

        if attach:
            return shell(
                (
                    "ssh",
                    "-ti",
                    project.config.ssh_key,
                    f"{project.config.ssh_user}@{ssh_host}",
                    "screen",
                    "-rS",
                    session,
                )
            )

        if not script:
            raise CommandError("Nothing to run")

        if script.startswith("@"):
            script = script[1:]
            return remote_exec(
                ssh_host=ssh_host,
                session=session,
                project=project,
                script=f'{script} {" ".join(additional_args)}',
                use_screen=use_screen,
            )

        if not os.path.exists(script):
            raise CommandError(f"{script} does not exists")

        script = os.path.abspath(script)
        if not context:
            context = os.path.dirname(script)

        remote_script = os.path.relpath(script, context)
        if remote_script.startswith("../"):
            raise CommandError("Script must be within context directory")

        remote_context = f"/home/{project.config.ssh_user}/opx-{session}"
        shell(
            (
                "rsync",
                "-rzue",
                f"ssh -i {project.config.ssh_key}",
                "--progress",
                f'{context.rstrip("/")}/',
                f"{project.config.ssh_user}@{ssh_host}:{remote_context}",
            )
        )

        if sync_only:
            return

        script = (
            f"cd {shlex.quote(remote_context)} && "
            f"chmod +x {shlex.quote(remote_script)} && "
            f'./{shlex.quote(remote_script)} {" ".join(map(shlex.quote, additional_args))}'
        ).strip()
        init_cmd = environment.context["RUN_INIT_CMD"]
        if init_cmd:
            script = f"{init_cmd} && {script}"

        script = f'bash -c "{script}"'
        shell_cmd = f"screen -L -S {session}" if use_screen else ""

        start_time = time.time()
        shell(
            (
                "ssh",
                "-ti",
                project.config.ssh_key,
                f"{project.config.ssh_user}@{ssh_host}",
                f"cd {shlex.quote(remote_context)}; {shell_cmd} {script}",
            )
        )

        if use_screen:
            elapsed_time = time.time() - start_time
            if elapsed_time < 5:
                logger.info("Screen session closed too soon. Retrieving logs...")
                shell(
                    (
                        "ssh",
                        "-ti",
                        project.config.ssh_key,
                        f"{project.config.ssh_user}@{ssh_host}",
                        f"less +G {shlex.quote(remote_context)}/screenlog.*",
                    )
                )
            else:
                logger.info(
                    "Screen session closed. To check logs:\n"
                    f'  strong-opx run {instance.hostname} "@less +G {shlex.quote(remote_context)}/screenlog.0" --no-screen'
                )



import argparse
import importlib.util
from typing import TYPE_CHECKING, Any

from strong_opx.exceptions import CommandError
from strong_opx.management.command import ProjectCommand

if TYPE_CHECKING:
    from strong_opx.project import Project


class Command(ProjectCommand):
    require_environment = False

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument("what", help="What to generate")

    def handle(self, project: "Project", what: str, **options: Any) -> None:
        try:
            module = importlib.import_module(f"strong_opx.codegen.generators.{what}")
            generator_class = getattr(module, "Generator")
        except ModuleNotFoundError:
            raise CommandError(f"Unknown generator: {what}")

        generator = generator_class(project)
        generator.generate()



import argparse
from typing import Any

from strong_opx.helm import HelmManager
from strong_opx.management.command import ProjectCommand
from strong_opx.platforms import KubernetesPlatform
from strong_opx.project import Environment


class Command(ProjectCommand):
    help_text = "Manage Helm Charts"
    allow_additional_args = True

    examples = [
        "strong-opx helm --project <project> --env <env> apply",
        "strong-opx helm --project <project> --env <env> apply --upgrade",
        "strong-opx helm --project <project> --env <env> apply --prune",
        "strong-opx helm --project <project> --env <env> -- uninstall <helm_chart_name>",
    ]

    def add_arguments(self, parser: argparse.ArgumentParser):
        super().add_arguments(parser)
        subparsers = parser.add_subparsers(title="operation", description="Operation to execute")

        apply = subparsers.add_parser("apply", help="Upgrade/install all charts")
        apply.add_argument("--prune", action="store_true", default=False, help="Uninstall removed packages")
        apply.add_argument("--upgrade", action="store_true", default=False, help="Upgrade installed packages")
        apply.add_argument("charts", metavar="chart", nargs="*", help="Specify charts to apply")
        apply.set_defaults(operation="apply")

    def handle(self, environment: Environment, additional_args: tuple[str, ...], operation: str = None, **options: Any):
        platform = environment.select_platform(KubernetesPlatform)

        manager = HelmManager(platform)
        if operation is None:
            manager.run(*additional_args)

        if operation == "apply":
            manager.apply(upgrade=options["upgrade"], charts=options["charts"], additional_args=additional_args)
            if options["prune"]:
                manager.prune(*additional_args)



import argparse
from typing import Any

from strong_opx.management.command import ProjectCommand
from strong_opx.project import Environment


class Command(ProjectCommand):
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument("nodes", nargs="+", metavar="node", help="What to deploy?")

    def handle(self, environment: Environment, nodes: list[str], **options: Any):
        environment.deploy(nodes)



import argparse
import logging
from typing import Any

from strong_opx.management.command import ProjectCommand
from strong_opx.platforms import GenericPlatform
from strong_opx.project import Environment, Project
from strong_opx.providers.compute import ComputeInstance, ComputeInstanceState

logger = logging.getLogger(__name__)


class Command(ProjectCommand):
    help_text = "Manage state of compute instance inside a project/environment"
    examples = [
        "strong-opx ec2 --project <project> --env <env> stop 10.0.0.1 ...",
        "strong-opx ec2 --project <project> --env <env> start primary ...",
        "strong-opx ec2 --project <project> --env <env> stop primary:2 ...",
    ]

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument("operation", help="Operation to execute", choices={"start", "stop", "restart", "status"})
        parser.add_argument("host_or_group", nargs="+", help="Hostname or host group or ip of instance")
        parser.add_argument("--wait", default=False, action="store_true", help="Wait for operation to complete")

    def handle(
        self,
        environment: Environment,
        project: Project,
        operation: str,
        host_or_group: list[str],
        wait: bool,
        **options: Any,
    ) -> None:
        platform = environment.select_platform(GenericPlatform)

        instances: list[ComputeInstance] = []
        for h in host_or_group:
            instances.extend(platform.resolve_instance(h))

        instance_states, running_instances, stopped_instances = self._sort_by_state(instances)

        if operation == "status":
            self._log_instance_states(instance_states, instances)
            return

        self._raise_for_any_invalid_states(instance_states)

        if operation in ("stop", "restart") and running_instances:
            # if restarting, we *need* to wait for instances to stop before we restart them
            wait_override = wait or operation == "restart"
            self._start_or_stop_instances(
                project=project, instance_ids=running_instances, operation="stop", wait=wait_override
            )

        if operation == "restart":
            stopped_instances.extend(running_instances)

        if operation in ("start", "restart") and stopped_instances:
            self._start_or_stop_instances(project=project, instance_ids=stopped_instances, operation="start", wait=wait)

    @staticmethod
    def _start_or_stop_instances(project: Project, instance_ids: list[str], operation: str, wait: bool):
        number_of_instances = len(instance_ids)
        logger.info(f'Executing "{operation}" operation for {number_of_instances} instance(s)...')

        if operation == "start":
            project.provider.start_compute_instance(instance_ids, wait=wait)
        elif operation == "stop":
            project.provider.stop_compute_instance(instance_ids, wait=wait)

        logger.info(f'"{operation}" operation complete for {number_of_instances} instance(s)')

    @staticmethod
    def _raise_for_any_invalid_states(instance_states: dict[ComputeInstance, ComputeInstanceState]):
        invalid_ips = []
        for ip, state in instance_states.items():
            if state not in (ComputeInstanceState.RUNNING, ComputeInstanceState.STOPPED):
                invalid_ips.append(f"(IP={ip}, State={state.value})")

        number_of_invalid_ips = len(invalid_ips)
        if number_of_invalid_ips > 0:
            raise RuntimeError(
                f'Found {number_of_invalid_ips} instance(s) with an invalid state: {", ".join(invalid_ips)}'
            )

    @staticmethod
    def _log_instance_states(
        instance_states: dict[ComputeInstance, ComputeInstanceState], instances: list[ComputeInstance]
    ):
        for instance in instances:
            logger.info(f"{instance.hostname} [{instance.instance_id}]: {instance_states[instance]}")

    @staticmethod
    def _sort_by_state(
        instances: list[ComputeInstance],
    ) -> tuple[dict[ComputeInstance, ComputeInstanceState], list[str], list[str]]:
        instance_states = {}
        stopped_instances = []
        running_instances = []

        for instance in instances:
            state = instance.current_state
            instance_states[instance] = state

            if state == ComputeInstanceState.RUNNING:
                running_instances.append(instance.instance_id)
            elif state == ComputeInstanceState.STOPPED:
                stopped_instances.append(instance.instance_id)

        return instance_states, running_instances, stopped_instances



from functools import cached_property
from typing import Any, Collection

from strong_opx.config.base import Config
from strong_opx.exceptions import ImproperlyConfiguredError

empty = object()


class HierarchicalConfig:
    missing_required_config_template = (
        "strong-opx is improperly configured. Configure {section}.{option} using\n\n"
        "  $ strong-opx config {section}.{option} <{placeholder}>"
    )

    def __init__(self, configs: Collection[Config]):
        self.configs = configs

    def get(self, section: str, option: str, fallback=None) -> Any:
        for config in self.configs:
            value = config.get(section, option, fallback=empty)
            if value is not empty:
                return value

        return fallback

    def get_required(self, section: str, option: str, placeholder="value") -> Any:
        value = self.get(section, option, fallback=empty)
        if value is empty:
            raise ImproperlyConfiguredError(
                self.missing_required_config_template.format(option=option, section=section, placeholder=placeholder)
            )

        return value

    @cached_property
    def ssh_user(self) -> str:
        return self.get_required("ssh", "user", placeholder="remote-login-user")

    @cached_property
    def ssh_key(self) -> str:
        return self.get_required("ssh", "key", placeholder="path-to-ssh-key")

    @cached_property
    def git_ssh_key(self) -> str:
        return self.get("git", "ssh.key", fallback=self.ssh_key)

    @cached_property
    def terraform_executable(self) -> str:
        return self.get("terraform", "executable", fallback="terraform")

    @cached_property
    def packer_executable(self) -> str:
        return self.get("packer", "executable", fallback="packer")

    @cached_property
    def ansible_playbook_executable(self) -> str:
        return self.get("ansible", "playbook.executable", fallback="ansible-playbook")

    @cached_property
    def docker_executable(self) -> str:
        return self.get("docker", "executable", fallback="docker")

    @cached_property
    def kubectl_executable(self) -> str:
        return self.get("kubectl", "executable", fallback="kubectl")
